import openml

import argparse
import json
import pickle
import time

from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.utils import shuffle

from stopit import ThreadingTimeout as Timeout, TimeoutException
import os

from genens import GenensClassifier
from genens.workflow.evaluate import SampleCrossValEvaluator, CrossValEvaluator
from genens.render.plot import export_plot
from genens.render.graph import create_graph


def evaluate_pipeline(pipe, X, y, out_dir):
    eval = CrossValEvaluator(cv_k=10)
    eval.fit(X, y)

    score = eval.score(pipe)  # accuracy score

    with open(out_dir + '/accuracy-score.txt', 'w+') as score_file:
        score_file.write('{}\n'.format(score))


def _create_dir_check(out_dir):
    try:
        os.mkdir(out_dir)
    except FileExistsError:
        # skip test
        print("Test directory {} already exists.".format(out_dir))
        return False
    except OSError as e:
        print("\nCannot create test directory.")
        raise e

    return True


def _log_evolution(fitted_clf, out_dir):
    # write logbook string representation to output dir
    with open(out_dir + '/logbook.txt', 'w+') as log_file:
        log_file.write(fitted_clf.logbook.__str__() + '\n')

    # evolution plot
    export_plot(fitted_clf, out_dir + '/result.png')

    # top 3 individual fitness values
    with open(out_dir + '/ind-fitness.txt', 'w+') as out_file:
        for i, ind in enumerate(fitted_clf.get_best_pipelines(as_individuals=True)[:3]):
            out_file.write('Individual {}: Score {}\n'.format(i, ind.fitness.values))

            # individual tree
            create_graph(ind, out_dir + '/graph{}.png'.format(i))

    # pickle top 3 best pipelines
    for i, pipe in enumerate(fitted_clf.get_best_pipelines()[:3]):
        with open(out_dir + '/pipeline{}.pickle'.format(i), 'wb') as pickle_file:
            pickle.dump(pipe, pickle_file, pickle.HIGHEST_PROTOCOL)


def _heuristic_sample_size(n_rows, n_cols):
    size = n_rows * n_cols

    # 'small' datasets
    if size < 10000:
        return None

    if n_rows < 1000:
        return 0.5

    if n_rows < 5000:
        return 0.25

    if n_rows < 10000:
        return 0.1

    if n_rows < 20000:
        return 0.05

    if n_cols < 10:
        return 0.02

    return 0.01


def _conditional_imput(X, categorical):
    categorical_id = [i for i, val in enumerate(categorical) if val]
    numerical_id = [i for i, val in enumerate(categorical) if not val]

    categorical_pipe = Pipeline(steps=[
        ('median_imputer', SimpleImputer(strategy='median'))
    ])

    numerical_pipe = Pipeline(steps=[
        ('mean_imputer', SimpleImputer(strategy='mean'))
    ])

    imputer = ColumnTransformer(
        transformers=[
            ('numerical', numerical_pipe, numerical_id),
            ('categorical', categorical_pipe, categorical_id),
        ]
    )

    return imputer


def run_task(task, out_dir, n_jobs=1, timeout=None, task_timeout=None):
    dataset = openml.datasets.get_dataset(task.dataset_id)

    X, y, categorical = dataset.get_data(
        target=dataset.default_target_attribute,
        return_categorical_indicator=True
    )

    imputer = _conditional_imput(X, categorical)
    X = imputer.fit_transform(X)

    # TODO hardcoded random state
    X, y = shuffle(X, y, random_state=42)

    sample_size = _heuristic_sample_size(X.shape[0], X.shape[1])

    if sample_size is not None:
        evaluator = SampleCrossValEvaluator(cv_k=5, timeout_s=timeout, per_gen=False,
                                            sample_size=sample_size)
    else:
        evaluator = CrossValEvaluator(cv_k=5, timeout_s=timeout)

    # TODO bad practice warning - hardcoded height
    clf = GenensClassifier(
        n_jobs=n_jobs,
        timeout=timeout,
        evaluator=evaluator,
        max_height=5
    )

    start_time = time.time()

    if task_timeout is not None:
        # timeout after task_timeout seconds
        try:
            with Timeout(task_timeout, swallow_exc=False):
                clf.fit(X, y)
        except TimeoutException:
            print('Timeout - task {} on dataset {}'.format(task.task_id, dataset.name))
            return
    else:
        # run without interrupt
        clf.fit(X, y)

    elapsed_time = time.time() - start_time

    # log time
    with open(out_dir + '/time.txt', 'w+') as time_file:
        time_file.write('Elapsed time: {}\n'.format(elapsed_time))

    _log_evolution(clf, out_dir)

    # run and log top 3 pipelines
    for i, pipe in enumerate(clf.get_best_pipelines()[:3]):
        pipe_dir = out_dir + '/run{}'.format(i)
        _create_dir_check(pipe_dir)

        evaluate_pipeline(pipe, X, y, pipe_dir)


def run_tests(task_ids=None, out_dir='.', n_jobs=1, timeout=None, task_timeout=None):
    benchmark_suite = openml.study.get_study('OpenML-CC18', 'tasks')

    for task_id in benchmark_suite.tasks:
        # skip other ids
        if task_ids is not None and task_id not in task_ids:
            continue

        task = openml.tasks.get_task(task_id)
        dataset = task.get_dataset()

        # create output directory or skip the task if it exists
        dataset_dir = '/{}'.format(dataset.name)
        if not _create_dir_check(out_dir + dataset_dir):
            continue

        run_task(task,
                 out_dir=out_dir + dataset_dir,
                 n_jobs=n_jobs,
                 timeout=timeout,
                 task_timeout=task_timeout)


if __name__ == '__main__':
    parser = argparse.ArgumentParser("Run OpenML task tests.")
    parser.add_argument('--config', type=str, help='Configuration file location.', required=True)
    parser.add_argument('--out', help='Output directory', required=True)

    args = parser.parse_args()

    with open(args.config) as cf:
        config = json.load(cf)

    kwargs = config.get('settings', {})
    task_ids = config.get('task_ids', None)

    run_tests(task_ids=task_ids,
              out_dir=args.out,
              **kwargs)



# -*- coding: utf-8 -*-
"""
Module for visualization of the evolution process.
"""
import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
import seaborn as sns


def export_plot(estimator, out_file):
    sns.set()

    gen = estimator.logbook.select("gen")
    score_max = estimator.logbook.chapters["score"].select("max")
    score_avg = estimator.logbook.chapters["score"].select("avg")

    test_max = estimator.logbook.chapters["test_score"].select("max")
    test_avg = estimator.logbook.chapters["test_score"].select("avg")

    test_all_zero = all(tt == 0.0 for tt in test_max) and \
        all(tt == 0.0 for tt in test_avg)

    fig = plt.figure()
    fig.suptitle("Evolution of population scores")

    ax1 = fig.add_subplot(111)

    ax1.set_xlabel("Generation")
    ax1.set_ylabel("Score")

    line1 = ax1.plot(gen, score_max, label='Maximum Score')
    line2 = ax1.plot(gen, score_avg, label='Average Score')

    if test_all_zero:
        lines = line1 + line2
    else:
        line3 = ax1.plot(gen, test_max, label='Maximum Test score')
        line4 = ax1.plot(gen, test_avg, label='Average Test score')

        lines = line1 + line2 + line3 + line4

    labels = [l.get_label() for l in lines]
    ax1.legend(lines, labels, loc="best")

    plt.savefig(out_file)
    plt.close()



# -*- coding: utf-8 -*-
"""
This module contains builtin types which extend the standard scikit-learn models.
"""


import math

from sklearn.base import BaseEstimator, TransformerMixin


class WeightedPipeline(BaseEstimator):
    """
    Pipeline that supports weights and can be therefore used as a base estimator.
    """
    def __init__(self, pipe):
        self.pipe = pipe

    def __repr__(self):
        return "Weighted pipeline: " + self.pipe.__repr__()

    def fit(self, X, y, sample_weight=None):
        return self.pipe.fit(X, y, predictor__sample_weight=sample_weight)

    def __getattr__(self, item):
        try:
            super().__getattribute__(item)
        except AttributeError:
            val = getattr(self.pipe, item)
            setattr(self, item, val)
            return val


class RelativeTransformer(BaseEstimator, TransformerMixin):
    """
    Transformer wrapper that enables the n_components or k to be specified by a fraction of the dataset size.
    """
    def __init__(self, transformer, feat_frac):
        self.transformer = transformer
        self.feat_frac = feat_frac

        if hasattr(self.transformer, 'k'):
            self.param_name = 'k'
        elif hasattr(self.transformer, 'n_components'):
            self.param_name = 'n_components'
        else:
            raise ValueError("Invalid feature transformer.")  # TODO specific

    def __repr__(self):
        return "Transformer ({}): ".format(self.feat_frac) + self.transformer.__repr__()

    def fit(self, X, y):
        if self.feat_frac is not None:
            param = int(math.ceil(self.feat_frac * X.shape[1]))

            if self.param_name == 'k':
                self.transformer.set_params(k=param)
            elif self.param_name == 'n_components':
                self.transformer.set_params(n_components=param)
            else:
                raise ValueError("Invalid feature transformer.")  # TODO specific

        self.transformer.fit(X, y)
        return self

    def __getattr__(self, item):
        try:
            super().__getattribute__(item)
        except AttributeError:
            val = getattr(self.transformer, item)
            setattr(self, item, val)
            return val



# -*- coding: utf-8 -*-

"""
This module contains the base estimator of Genens.
"""

import json
import logging
import logging.config
import warnings
from queue import Queue

import numpy as np
import os

from genens.gp.tree import gen_tree
from genens.gp.evolution import gen_individual
from genens.gp.mutation import mutate_subtree, mutate_gradual_hillclimbing
from genens.gp.mutation import perform_hillclimbing, mutate_args
from genens.gp.mutation import mutate_node_swap
from genens.gp.crossover import crossover_one_point
from genens.gp.evolution import ea_run

from genens.log_utils import set_log_handler

from genens.workflow.evaluate import CrossValEvaluator, TrainTestEvaluator, default_score
from genens.workflow.model_creation import create_workflow

from deap import base, tools
from functools import partial
from joblib import delayed, Parallel
from multiprocessing import Manager
from logging.handlers import QueueListener, QueueHandler

from sklearn.base import BaseEstimator, is_classifier
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels


file_dir = os.path.dirname(os.path.abspath(__file__))
DEFAULT_LOGGING_CONFIG = file_dir + '/.logging_config.json'


class GenensBase(BaseEstimator):
    def __init__(self, config, n_jobs=1, cx_pb=0.5, mut_pb=0.3, mut_args_pb=0.9,
                 mut_node_pb=0.9, scorer=None, pop_size=100,
                 mut_multiple_args=False, mut_multiple_nodes=False,
                 n_gen=15, hc_repeat=0, hc_keep_last=False, hc_mut_pb=0.2, hc_n_nodes=3,
                 weighted=True, use_groups=True, max_height=None,
                 max_arity=None, timeout=None, evaluator=None,
                 log_path=None, max_evo_seconds=None):
        """
        Creates a new Genens estimator.

        :param GenensConfig config: Configuration of Genens.
        :param n_jobs: The n_jobs parameter for the process of evolution.
        :param cx_pb: Crossover probability.
        :param mut_pb: Subtree mutation probability.
        :param mut_args_pb: Argument mutation probability.
        :param mut_node_pb: Point (node) mutation probability.
        :param scorer: Scorer to be used during all evaluation (must comply to scikit-learn scorer API).
        :param pop_size: Population size.
        :param n_gen: Number of generations.
        :param hc_repeat: Number of hill-climbing iteration. If set to 0, hill-climbing is not performed.
        :param hc_keep_last:
            Whether the last individual should be mutated if the hill-climbing did not find a better individual.

        :param hc_mut_pb: Probability of additional hillclimbing performed on offspring.
        :param hc_n_nodes: Number of nodes mutated during hillclimbing (`hc_repeat` times, `hc_keep_last` has effect)


        :param weighted: Determines whether the selection of nodes is weighted (according to groups).

        :param bool use_groups: If false, primitive groups are ignored.
        :param max_height: Maximum height of tree individuals.
        :param max_arity: Maximum arity of all nodes.
        :param timeout: Timeout for a single method evaluation.
        :param evaluator: Evaluator to be used (see genens.worflow.evaluate)

        """

        # accept config/load default config
        self.config = config
        if max_height is not None:
            config.max_height = max_height

        if max_arity is not None:
            config.max_arity = max_arity

        self.n_jobs = n_jobs

        self.cx_pb = cx_pb
        self.mut_pb = mut_pb
        self.mut_args_pb = mut_args_pb
        self.mut_node_pb = mut_node_pb

        self.pop_size = pop_size
        self.n_gen = n_gen

        self.mut_multiple_args = mut_multiple_args
        self.mut_multiple_nodes = mut_multiple_nodes

        self.hc_repeat = hc_repeat
        self.hc_keep_last = hc_keep_last
        self.hc_n_nodes = hc_n_nodes
        self.hc_mut_pb = hc_mut_pb
        self.weighted = weighted
        self.use_groups = use_groups

        self.scorer = scorer

        self._timeout = timeout
        self._fitness_evaluator = evaluator if evaluator is not None \
            else CrossValEvaluator(timeout_s=timeout)
        self._fitness_evaluator.timeout = timeout

        self.max_evo_seconds = max_evo_seconds

        self.test_evaluator = None

        self.pareto = tools.ParetoFront()
        self.fitted_wf = None
        self._population = None

        self._log_path = log_path
        self._logging_config = DEFAULT_LOGGING_CONFIG
        self._setup_debug_logging()
        self._setup_stats_logging()
        self._setup_toolbox()

    def _setup_arg_mut(self):
        mut_func = partial(mutate_args, self.config,
                           multiple_nodes=self.mut_multiple_nodes,
                           multiple_args=self.mut_multiple_args)

        self._toolbox.register("gradual_hillclimbing", mutate_gradual_hillclimbing,
                               self._toolbox, self.config,
                               hc_repeat=self.hc_repeat, keep_last=self.hc_keep_last)
        self._toolbox.register("mutate_args", mut_func)

    def _setup_toolbox(self):
        self._toolbox = base.Toolbox()

        self._toolbox.register("individual", gen_tree, self.config,
                               weighted=self.weighted, use_groups=self.use_groups)

        pop_func = partial(gen_individual, self._toolbox, self.config)
        self._toolbox.register("population", tools.initRepeat, list, pop_func)

        self._toolbox.register("map", _map_parallel)

        self._toolbox.register("select", tools.selNSGA2)
        self._toolbox.register("mutate_subtree", mutate_subtree, self._toolbox)
        self._setup_arg_mut()
        self._toolbox.register("mutate_node_swap", mutate_node_swap, self.config)
        self._toolbox.register("cx_one_point", crossover_one_point)

        self._toolbox.register("compile", self._compile_pipe)
        self._toolbox.register("evaluate", self._eval_tree_individual)

        self._toolbox.register("update", self._update_population)
        self._toolbox.register("log_setup", self._setup_child_logging)

    def _load_log_config(self):
        if self._logging_config is not None:
            with open(self._logging_config, 'r') as f:
                return json.load(f)

    def _setup_debug_logging(self):
        config = self._load_log_config()
        logging.config.dictConfig(config)
        self.log_format = config['format']

        if self.n_jobs != 1:
            mp_manager = Manager()
            self._log_queue = mp_manager.Queue()
        else:
            self._log_queue = Queue()

        handl = QueueHandler(self._log_queue)
        logger = logging.getLogger("genens")

        logger.addHandler(handl)

    def _setup_child_logging(self):
        config = self._load_log_config()
        logging.config.dictConfig(config)

        handl = QueueHandler(self._log_queue)
        logger = logging.getLogger("genens")
        logger.addHandler(handl)
        return handl

    def _setup_stats_logging(self):
        score_stats = tools.Statistics(lambda ind: ind.fitness.values[0])
        test_stats = tools.Statistics(lambda ind: self._compute_test(ind))

        self._mstats = tools.MultiStatistics(score=score_stats, test_score=test_stats)

        self._mstats.register("avg", np.mean)
        self._mstats.register("std", np.std)
        self._mstats.register("min", np.min)
        self._mstats.register("max", np.max)

        self.logbook = tools.Logbook()

        self.logbook.header = "gen", "score", "test_score"
        self.logbook.chapters["score"].header = "min", "avg", "max", "std"
        self.logbook.chapters["test_score"].header = "min", "avg", "max", "std"

    def _update_population(self, population, gen_i):
        self._fitness_evaluator.reset()
        self._log_pop_stats(population, gen_i)

    def _compile_pipe(self, ind):
        if ind.compiled_pipe is not None:
            return ind.compiled_pipe

        return create_workflow(ind, self.config.func_config)

    def _eval_tree_individual(self, gp_tree):
        wf = self._toolbox.compile(gp_tree)
        return self._fitness_evaluator.score(wf, scorer=self.scorer)

    @property
    def can_log_score(self):
        return self.test_evaluator is not None

    def setup_test_stats(self, train_X, train_y, test_X, test_y):
        self.test_evaluator = TrainTestEvaluator(test_X, test_y, timeout_s=self._timeout)
        self.test_evaluator.fit(train_X, train_y)

    def _compute_test(self, ind):
        if not self.can_log_score:
            return 0.0  # TODO, warn

        if ind.test_stats is not None:
            return ind.test_stats

        wf = self._toolbox.compile(ind)
        res = self.test_evaluator.score(wf, scorer=self.scorer)

        ind.test_stats = res
        return res if res is not None else 0.0  # TODO

    def _log_pop_stats(self, population, gen_id):
        self.pareto.update(population)

        record = self._mstats.compile(population)
        self.logbook.record(gen=gen_id, **record)

    def get_best_pipelines(self, as_individuals=False):
        check_is_fitted(self, 'is_fitted_')

        if as_individuals:
            [self._compute_test(ind) for ind in self.pareto]
            return self.pareto
        else:
            return list(map(self._toolbox.compile, self.pareto))

    def fit(self, train_X, train_y, verbose=1):
        train_X, train_y = check_X_y(train_X, train_y, accept_sparse=True)

        if is_classifier(self):
            self.classes_ = unique_labels(train_y)

        self._fitness_evaluator.fit(train_X, train_y)
        self.pareto.clear()

        self._population = self._toolbox.population(n=self.pop_size)

        # handlers do not work well with pickling (required for multiprocessing)
        formatter = logging.Formatter(fmt=self.log_format)
        if self._log_path is not None:
            handl = logging.FileHandler(self._log_path)
        else:
            handl = logging.StreamHandler()

        handl.setFormatter(formatter)
        log_queue_listener = QueueListener(self._log_queue, handl)

        try:
            log_queue_listener.start()
            ea_run(self._population, self._toolbox, n_gen=self.n_gen, pop_size=self.pop_size, cx_pb=self.cx_pb,
                   mut_pb=self.mut_pb, hc_mut_pb=self.hc_mut_pb, hc_n_nodes=self.hc_n_nodes,
                   mut_args_pb=self.mut_args_pb, mut_node_pb=self.mut_node_pb, n_jobs=self.n_jobs,
                   timeout=self.max_evo_seconds, verbose=verbose)

        finally:
            # close local logging handlers
            log_queue_listener.stop()
            handl.close()
            del log_queue_listener
            del handl

        if not len(self.pareto):
            # try to get individuals that were evaluated before time limit end
            evaluated_inds = [ind for ind in self._population if ind.fitness.valid]

            if not len(evaluated_inds):
                warnings.warn("The algorithm did not have enough time to evaluate first generation and was not fitted.")
                return self

            self.pareto.update(evaluated_inds)

        tree = self.pareto[0]
        self.fitted_wf = self._toolbox.compile(tree)
        self.fitted_wf.fit(train_X, train_y)

        self.is_fitted_ = True
        return self

    def predict(self, test_X):
        test_X = check_array(test_X, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')

        # TODO clf/regr specific

        res = self.fitted_wf.predict(test_X)
        return res

    def score(self, test_X, test_y):
        test_X, test_y = check_X_y(test_X, test_y, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')

        if self.scorer is not None:
            s = self.scorer(self.fitted_wf, test_X, test_y)
        else:
            s = default_score(self.fitted_wf, test_X, test_y)

        return s


def _map_parallel(func, population, n_jobs=1, **kwargs):
    return Parallel(n_jobs=n_jobs, **kwargs)(delayed(func)(ind) for ind in population)



import argparse
import numpy as np
import random

from genens import GenensClassifier
from genens.workflow.evaluate import SampleCrossValEvaluator, CrossValEvaluator
from genens.tests.load_datasets import load_dataset


def run_test(kwarg_dict):
    random_state = kwarg_dict.pop('random_state')
    np.random.seed(random_state)
    random.seed(random_state)

    dataset = kwarg_dict.pop('dataset')
    X, y = load_dataset(dataset)

    cv_k = kwarg_dict.pop('cv_k')
    size = kwarg_dict.pop('size')

    if size == 1.0:
        evaluator = CrossValEvaluator(cv_k=cv_k)
    else:
        evaluator = SampleCrossValEvaluator(cv_k=cv_k, sample_size=size, per_gen=True)

    clf = GenensClassifier(
        evaluator=evaluator,
        **kwarg_dict
    )

    clf.fit(X, y, verbose=2)
    return clf, X, y


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', required=True)
    parser.add_argument('--random_state', default=42, type=int)

    parser.add_argument('--n_jobs', default=-1, type=int)
    parser.add_argument('--pop_size', default=200, type=int)
    parser.add_argument('--n_gen', default=15, type=int)

    parser.add_argument('--max_height', default=4, type=int)
    parser.add_argument('--max_arity', default=3, type=int)

    parser.add_argument('--weighted', action='store_true')
    parser.add_argument('--use_groups', action='store_true')

    parser.add_argument('--cx_pb', default=0.5, type=float)
    parser.add_argument('--mut_pb', default=0.3, type=float)
    parser.add_argument('--mut_args_pb', default=0.6, type=float)
    parser.add_argument('--mut_node_pb', default=0.3, type=float)

    parser.add_argument('--cv_k', default=3, type=int)
    parser.add_argument('--size', default=0.25, type=float)
    parser.add_argument('--timeout', default=240, type=int)
    parser.add_argument('--log_path', default=None)

    args = parser.parse_args()

    run_test(vars(args))



# -*- coding: utf-8 -*-

"""
This module provides methods which convert the tree representation to pipelines.
Every method operates on specific nodes along with its (already converted) child nodes.
"""


import inspect
from functools import reduce

from sklearn.pipeline import Pipeline, make_union

from .builtins import WeightedPipeline, RelativeTransformer


def create_workflow(gp_tree, config_dict):
    """
    Creates a pipeline (workflow) from the tree.

    :param gp_tree: Tree to be converted.
    :param config_dict: Configuration to be used for conversion.
    :return: A scikit-learn pipeline.
    """
    def wf_step_from_node(node, child_list):
        return config_dict[node.name](child_list, node.obj_kwargs)

    return gp_tree.run_tree(wf_step_from_node)


def create_stacking(ens_cls, const_kwargs, child_list, evolved_kwargs):
    if not len(child_list) or len(child_list) < 2:
        raise ValueError("Not enough estimators provided to the ensemble.")

    estimator_list = child_list[:-1]
    estimator_list = [(f'clf{i}', clf) for i, clf in enumerate(estimator_list)]
    final_estimator = child_list[-1]

    return ens_cls(estimators=estimator_list,
                   final_estimator=final_estimator,
                   **const_kwargs, **evolved_kwargs)


def create_ensemble(ens_cls, const_kwargs, child_list, evolved_kwargs):
    """
    Creates an ensemble with its children set as base-learners.

    :param ens_cls: Function that creates the ensemble
    :param const_kwargs: Keyword arguments which do not change during the evolution.
    :param child_list: List of converted child nodes.
    :param evolved_kwargs: Keyword arguments which are set during the evolution process.
    :return: A new ensemble.
    """
    if not len(child_list):
        raise ValueError("No base estimator provided to the ensemble.")  # TODO specific

    if 'base_estimator' in inspect.signature(ens_cls).parameters:
        if len(child_list) != 1:
            raise ValueError("Incorrect number of base estimators.")  # TODO specific

        ens = ens_cls(**const_kwargs, **evolved_kwargs, base_estimator=child_list[0])
    elif 'estimators' in inspect.signature(ens_cls).parameters:

        est_names = ['clf' + str(i) for i in range(0, len(child_list))]
        est_list = list(zip(est_names, child_list))

        ens = ens_cls(**const_kwargs, **evolved_kwargs, estimators=est_list)
    else:
        raise ValueError("Invalid ensemble - missing constructor parameters.")  # TODO specific

    return ens


def create_estimator(est_cls, const_kwargs, child_list, evolved_kwargs):
    """
    Creates an estimator.

    :param est_cls: Function that creates the estimator.
    :param const_kwargs: Keyword arguments which do not change during the evolution.
    :param child_list: List of converted child nodes - should me empty.
    :param evolved_kwargs: Keyword arguments which are set during the evolution process.
    :return: A new estimator.
    """

    if len(child_list) > 0:
        raise ValueError("Estimator cannot have sub-estimators.")  # TODO specific

    if 'feat_frac' in evolved_kwargs.keys():
        feat_frac = evolved_kwargs['feat_frac']
        evolved_kwargs = {key: val for key, val in evolved_kwargs.items()
                          if key != 'feat_frac'}

        est = est_cls(**const_kwargs, **evolved_kwargs)
        return RelativeTransformer(est, feat_frac)

    return est_cls(**const_kwargs, **evolved_kwargs)


def create_transform_list(child_list, evolved_kwargs):
    """
    Creates a transformer chain.

    :param child_list: List of transformers.
    :param evolved_kwargs: Keyword arguments, not used.
    :return:
    """
    return child_list


def create_empty_data(child_list, evolved_kwargs):
    """
    Creates an empty transformer list.

    :param child_list: Child list, must be empty.
    :param evolved_kwargs: Keyword arguments, not used.
    :return:
    """
    if len(child_list) > 0:
        raise ValueError("This can be assigned only to terminals.")  # TODO specific

    return []


def create_pipeline(child_list, evolved_kwargs):
    """
    Creates a pipeline from the child list.

    :param child_list: Should contain either an estimator or an estimator along with a transformer chain.
    :param evolved_kwargs: Keyword arguments, not used.
    :return: A scikit-learn pipeline.
    """
    if len(child_list) > 2 or not len(child_list):
        raise ValueError("Invalid child list for pipeline.")  # TODO specific

    predictor = child_list[0]

    if len(child_list) > 1:
        if isinstance(child_list[1], list):
            step_names = ['step' + str(i) for i in range(0, len(child_list[1]))]
            steps = list(zip(step_names, child_list[1]))
        else:
            steps = [('step', child_list[1])]
        steps.append(('predictor', predictor))

        pipe = Pipeline(steps=steps, **evolved_kwargs)
    else:
        pipe = Pipeline(steps=[('predictor', predictor)], **evolved_kwargs)

    if 'sample_weight' in inspect.signature(predictor.fit).parameters:
        wp = WeightedPipeline(pipe)
        return wp

    return pipe


def create_data_union(child_list, evolved_kwargs):
    """
    Creates a data union from the child list.
    :param child_list: List of transformers.
    :param evolved_kwargs: Keyword arguments of the union.
    :return: A scikit-learn FeatureUnion
    """
    if not len(child_list):
        raise ValueError("No base estimator provided to the feature union.")  # TODO specific

    def add_or_concat(res, child):
        if isinstance(child, list):
            return res + child
        res.append(child)
        return res

    reduced = reduce(add_or_concat, child_list, [])
    if not len(reduced):
        return []

    return [make_union(*reduced, **evolved_kwargs)]



# -*- coding: utf-8 -*-

from ..config.utils import ensemble_func
from ..config.utils import estimator_func
from ..config.utils import get_default_config
from ..config.utils import ensemble_primitive
from ..config.utils import predictor_primitive
from ..config.utils import predictor_terminal
from ..config.utils import transformer_primitive

from sklearn import decomposition
from sklearn import feature_selection
from sklearn import preprocessing

from sklearn import svm
from sklearn import linear_model
from sklearn import discriminant_analysis
from sklearn import neural_network
from sklearn import naive_bayes
from sklearn import tree
from sklearn import neighbors

from sklearn import ensemble


def clf_config(group_weights=None):
    """
    Creates Genens configuration for classifier problems.

    :param dict group_weights: Dictionary of group weights.
           If not provided, default setting is used.

    :return GenensConfig: Genens configuration for classifier problems.
    """

    # FUNCTIONS
    ensembles_func = {
        'voting': ensemble_func(ensemble.VotingClassifier)
    }

    clf_func = {
        'ada': estimator_func(ensemble.AdaBoostClassifier),
        'bagging': estimator_func(ensemble.BaggingClassifier),
        "KNeighbors": estimator_func(neighbors.KNeighborsClassifier),
        "LinearSVC": estimator_func(svm.LinearSVC),
        "SVC": estimator_func(svm.SVC),
        "logR": estimator_func(linear_model.LogisticRegression),
        "Perceptron": estimator_func(linear_model.Perceptron),
        "SGD": estimator_func(linear_model.SGDClassifier),
        "PAC": estimator_func(linear_model.PassiveAggressiveClassifier),
        "LDA": estimator_func(discriminant_analysis.LinearDiscriminantAnalysis),
        "QDA": estimator_func(discriminant_analysis.QuadraticDiscriminantAnalysis),
        "MLP": estimator_func(neural_network.MLPClassifier),
        "gaussianNB": estimator_func(naive_bayes.GaussianNB),
        "DT": estimator_func(tree.DecisionTreeClassifier),
        "gradBoosting": estimator_func(ensemble.GradientBoostingClassifier),
        "randomForest": estimator_func(ensemble.RandomForestClassifier),
        "extraTrees": estimator_func(ensemble.ExtraTreesClassifier)
    }

    transform_func = {
            "NMF": estimator_func(decomposition.NMF),
            "FA": estimator_func(decomposition.FactorAnalysis),
            "FastICA": estimator_func(decomposition.FastICA),
            "PCA": estimator_func(decomposition.PCA),
            # "KernelPCA" : estimator_func(decomposition.KernelPCA),
            "kBest": estimator_func(feature_selection.SelectKBest),
            "MaxAbsScaler": estimator_func(preprocessing.MaxAbsScaler),
            "MinMaxScaler":estimator_func(preprocessing.MinMaxScaler),
            "Normalizer": estimator_func(preprocessing.Normalizer),
            "StandardScaler": estimator_func(preprocessing.StandardScaler)
    }

    ensemble_kwargs = {
        'voting': {
            'voting': ['soft']
            #'voting': ['hard', 'soft']
        }
    }

    clf_kwargs = {
        'ada': {
            'n_estimators': [5, 10, 50, 100, 200],
            'algorithm': ['SAMME', 'SAMME.R']
        },
        'bagging': {
            'n_estimators': [5, 10, 50, 100, 200]
        },
        'KNeighbors': {
            'n_neighbors': [1, 2, 5],
            'algorithm': ['auto', 'ball_tree', 'kd_tree', 'brute']
        },
        'LinearSVC': {
            'loss': ['hinge', 'squared_hinge'],
            'penalty': ['l1', 'l2'],
            'C': [0.1, 0.5, 1.0, 2, 5, 10, 15],
            'tol': [0.0001, 0.001, 0.01]
        },
        'SVC': {
            'C': [0.1, 0.5, 1.0, 2, 5, 10, 15],
            'gamma': ['scale', 0.0001, 0.001, 0.01, 0.1, 0.5],
            'tol': [0.0001, 0.001, 0.01],
            'probability': [True, False]
        },
        'logR': {
            'penalty': ['l1', 'l2'],
            'C': [0.1, 0.5, 1.0, 2, 5, 10, 15],
            'tol': [0.0001, 0.001, 0.01],
            'solver': ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga']
        },
        'Perceptron': {
            'penalty': ['None', 'l2', 'l1', 'elasticnet'],
            'max_iter': [1, 2, 5, 10, 100],
            'alpha': [0.0001, 0.001, 0.01]
        },
        'SGD': {
            'penalty': ['none', 'l2', 'l1', 'elasticnet'],
            'loss': ['hinge', 'log', 'modified_huber', 'squared_hinge', 'perceptron'],
            'max_iter': [10, 100, 200],
            'tol': [0.0001, 0.001, 0.01],
            'alpha': [0.0001, 0.001, 0.01],
            'l1_ratio': [0, 0.15, 0.5, 1],
            'epsilon': [0.01, 0.05, 0.1, 0.5],
            'learning_rate': ['constant', 'optimal'],
            'eta0': [0.01, 0.1, 0.5],  # wild guess
            'power_t': [0.1, 0.5, 1, 2]  # dtto
        },
        'PAC': {
            'loss': ['hinge', 'squared_hinge'],
            'C': [0.1, 0.5, 1.0, 2, 5, 10, 15]
        },
        'LDA': {
            'solver': ['lsqr', 'eigen'],
            'shrinkage': [None, 'auto', 0.1, 0.5, 1.0]
        },
        'QDA': {
            'reg_param': [0.0, 0.1, 0.5, 1],
            'tol': [0.0001, 0.001, 0.01]
        },
        'MLP': {
            'activation': ['identity', 'logistic', 'relu'],
            'solver': ['lbfgs', 'sgd', 'adam'],
            'alpha': [0.0001, 0.001, 0.01],
            'learning_rate': ['constant', 'invscaling', 'adaptive'],
            'tol': [0.0001, 0.001, 0.01],
            'max_iter': [10, 100, 200],
            'learning_rate_init': [0.0001, 0.001, 0.01],
            'power_t': [0.1, 0.5, 1, 2],
            'momentum': [0.1, 0.5, 0.9],
            'hidden_layer_sizes': [(100,), (50,), (20,), (10,)]
        },
        'DT': {
            'criterion': ['gini', 'entropy'],
            'max_features': [0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'max_depth': [1, 2, 5, 10, 15, 25, 50, 100],
            'min_samples_split': [2, 5, 10, 20],
            'min_samples_leaf': [1, 2, 5, 10, 20]
        },
        'gaussianNB': {},
        'gradBoosting': {
            'loss': ['deviance', 'exponential'],
            'n_estimators': [20, 50, 100, 200],
            'subsample': [0.3, 0.5, 0.75, 1.0]
            # TODO
        },
        'randomForest': {
            'n_estimators': [10, 50, 100, 150, 200]
            # TODO
        },
        'extraTrees': {
            'n_estimators': [10, 50, 100, 150, 200]
        }
    }

    transform_kwargs = {
        'NMF': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'solver': ['cd', 'mu']
        },
        'FA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
        },
        'FastICA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],

        },
        'PCA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'whiten': [False, True],
        },
        # 'KernelPCA': {
        #    'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
        #    'kernel': ['linear', 'poly', 'rbf', 'sigmoid', 'cosine']
        # },
        'kBest': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'score_func': [feature_selection.chi2, feature_selection.f_classif]
        },
        'MaxAbsScaler': {},
        'MinMaxScaler': {},
        "Normalizer": {},
        "StandardScaler": {}
    }

    func_dict = {**ensembles_func, **clf_func, **transform_func}
    kwargs_dict = {**ensemble_kwargs, **clf_kwargs, **transform_kwargs}

    if group_weights is None:
        group_weights = {
            "pipeline": 1.0,
            # "union": 0.3,
            "prepro": 1.0,
            "ensemble": 0.5,
            "predictor": 1.0,
            "transform": 1.0
        }

    # add to config
    config = get_default_config()
    config.add_functions_args(func_dict, kwargs_dict)

    # PRIMITIVES

    # ensemble config
    config.add_primitive(ensemble_primitive('voting', (2, 'n')))

    # classifier config
    config.add_primitive(predictor_primitive('ada'))
    config.add_primitive(predictor_primitive('bagging'))
    config.add_primitive(predictor_primitive("KNeighbors"))
    config.add_primitive(predictor_primitive("LinearSVC"))
    config.add_primitive(predictor_primitive("SVC"))
    config.add_primitive(predictor_primitive("logR"))
    config.add_primitive(predictor_primitive("Perceptron"))
    config.add_primitive(predictor_primitive("SGD"))
    config.add_primitive(predictor_primitive("PAC"))
    config.add_primitive(predictor_primitive("gaussianNB"))
    config.add_primitive(predictor_primitive("DT"))
    config.add_primitive(predictor_primitive("gradBoosting"))
    config.add_primitive(predictor_primitive("randomForest"))
    config.add_primitive(predictor_primitive("extraTrees"))

    # terminals used only as leaves
    config.add_primitive(predictor_terminal("KNeighbors"), term_only=True)
    config.add_primitive(predictor_terminal("LinearSVC"), term_only=True)
    config.add_primitive(predictor_terminal("SVC"), term_only=True)
    config.add_primitive(predictor_terminal("logR"), term_only=True)
    config.add_primitive(predictor_terminal("Perceptron"), term_only=True)
    config.add_primitive(predictor_terminal("SGD"), term_only=True)
    config.add_primitive(predictor_terminal("PAC"), term_only=True)
    config.add_primitive(predictor_terminal("gaussianNB"), term_only=True)
    config.add_primitive(predictor_terminal("DT"), term_only=True)
    config.add_primitive(predictor_terminal("gradBoosting"), term_only=True)
    config.add_primitive(predictor_terminal("randomForest"), term_only=True)
    config.add_primitive(predictor_terminal("extraTrees"), term_only=True)

    # transformer config

    # can be logically ordered via default cData node
    config.add_primitive(transformer_primitive("NMF", 'featsel'))
    config.add_primitive(transformer_primitive("FA", 'featsel'))
    config.add_primitive(transformer_primitive("FastICA", 'featsel'))
    config.add_primitive(transformer_primitive("PCA", 'featsel'))
    #config.add_primitive(transformer_primitive("KernelPCA", 'featsel'))
    config.add_primitive(transformer_primitive("kBest", 'featsel'))

    config.add_primitive(transformer_primitive("MaxAbsScaler", 'scale'))
    config.add_primitive(transformer_primitive("MinMaxScaler", 'scale'))
    config.add_primitive(transformer_primitive("Normalizer", 'scale'))
    config.add_primitive(transformer_primitive("StandardScaler", 'scale'))

    # terminals used only as leaves
    config.add_primitive(transformer_primitive("NMF", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("FA", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("FastICA", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("PCA", 'data'), term_only=True)
    # config.add_primitive(transformer_primitive("KernelPCA", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("kBest", 'data'), term_only=True)

    config.add_primitive(transformer_primitive("MaxAbsScaler", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("MinMaxScaler", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("Normalizer", 'data'), term_only=True)
    config.add_primitive(transformer_primitive("StandardScaler", 'data'), term_only=True)

    config.group_weights = group_weights  # checks for missing values according to config terminals

    return config



# -*- coding: utf-8 -*-

import openml
import os
import pandas as pd
from functools import partial

from sklearn.datasets import load_iris, load_digits, load_wine, load_breast_cancer, fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

from sklearn.preprocessing import LabelEncoder


def load_mnist(split_validation=False, random_state=None, test_size=None):
    features, target = fetch_openml('mnist_784', return_X_y=True)

    if split_validation:
        train_X, test_X, train_y, test_y = train_test_split(features, target, test_size=test_size,
                                                            random_state=random_state)
        return train_X, train_y, test_X, test_y

    features, target = shuffle(features, target, random_state=random_state)
    return features, target


def load_magic(split_validation=False, random_state=None, test_size=None):
    dir_path = os.path.dirname(__file__)

    filename = dir_path + '/magic.csv'

    data = pd.read_csv(filename, sep=',')
    data = shuffle(data)

    features = data[data.columns[:-1]]
    target = data[data.columns[-1]]
    le = LabelEncoder()

    ix = target.index
    target = pd.Series(le.fit_transform(target), index=ix)

    if split_validation:
        train_X, test_X, train_y, test_y = train_test_split(features, target, test_size=test_size,
                                                            random_state=random_state)
        return train_X, train_y, test_X, test_y

    features, target = shuffle(features, target, random_state=random_state)
    return features, target


def load_wilt(split_validation=False, random_state=None, test_size=None):
    use_original_test = split_validation and test_size is None

    dir_path = os.path.dirname(__file__)

    filename = dir_path + '/wilt-train.csv'
    data = pd.read_csv(filename, sep=',')

    test_filename = dir_path + '/wilt-test.csv'
    test_data = pd.read_csv(test_filename, sep=',')

    # concatenate both sets to create the full set
    if not use_original_test:
        data = pd.concat([data, test_data])

    features = data[data.columns[1:]]
    target = data[data.columns[0]]
    le = LabelEncoder()

    ix = target.index
    target = pd.Series(le.fit_transform(target), index=ix)

    # repeat process for validation set, return the original test
    if use_original_test:
        test_X = test_data[test_data.columns[1:]]
        test_y = test_data[test_data.columns[0]]
        le = LabelEncoder()

        ix = test_y.index
        test_y = pd.Series(le.fit_transform(test_y), index=ix)

        return features, target, test_X, test_y

    # choose a different test set
    if split_validation and test_size is not None:
        train_X, test_X, train_y, test_y = train_test_split(features, target,
                                                            test_size=test_size,
                                                            random_state=random_state)
        return train_X, train_y, test_X, test_y

    features, target = shuffle(features, target, random_state=random_state)
    return features, target


def load_from_sklearn(load_func, split_validation=False, random_state=None, test_size=None):
    features, target = load_func(return_X_y=True)

    if split_validation:
        train_X, test_X, train_y, test_y = train_test_split(features, target,
                                                            test_size=test_size,
                                                            random_state=random_state)
        return train_X, train_y, test_X, test_y

    features, target = shuffle(features, target, random_state=random_state)
    return features, target


def load_from_openml(dataset_name, split_validation=False, random_state=None, test_size=None):
    all_datasets = openml.datasets.list_datasets()

    dataset_id = None
    for k, v in all_datasets.items():
        if v['name'] == dataset_name:
            dataset_id = k
            break

    if dataset_id is None:
        raise ValueError("Invalid dataset name.")

    dataset = openml.datasets.get_dataset(dataset_id)

    features, target, _, _ = dataset.get_data(target=dataset.default_target_attribute)

    if split_validation:
        train_X, test_X, train_y, test_y = train_test_split(features, target,
                                                            test_size=test_size,
                                                            random_state=random_state)
        return train_X, train_y, test_X, test_y

    features, target = shuffle(features, target, random_state=random_state)
    return features, target


#
load_functions = {}


"""
load_functions = {
    'iris': partial(load_from_sklearn, load_iris),
    'digits': partial(load_from_sklearn, load_digits),
    'wine': partial(load_from_sklearn, load_wine),
    'breast_cancer': partial(load_from_sklearn, load_breast_cancer),
    'wilt': load_wilt,
    'magic': load_magic,
    'mnist': load_mnist
}
"""


def load_dataset(dataset_name, split_validation=False, random_state=None, test_size=None):
    """
    # TODO
    classification: 'wine' 'breast_cancer' 'digits' 'iris' ...
    regression: 'diabetes' 'boston' 'california (fetch)' ...
    """

    if dataset_name in load_functions:
        return load_functions[dataset_name](split_validation=split_validation,
                                            random_state=random_state,
                                            test_size=test_size)
    else:
        return load_from_openml(dataset_name,
                                split_validation=split_validation,
                                random_state=random_state,
                                test_size=test_size)





# -*- coding: utf-8 -*-

"""
This module provides evaluators that can be used as fitness evaluators in Genens.
"""

from abc import ABC, abstractmethod
from functools import wraps

import logging
import numpy as np

from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score
from sklearn.metrics import accuracy_score

from stopit import ThreadingTimeout as Timeout, TimeoutException

import time
import warnings


def timeout(fn):
    @wraps(fn)
    def with_timeout(self, workflow, *args, **kwargs):
        if not hasattr(self, 'timeout') or self.timeout is None:
            return fn(self, workflow, *args, **kwargs)

        try:
            with Timeout(self.timeout, swallow_exc=False):
                res = fn(self, workflow, *args, **kwargs)
        except TimeoutException:
            logger = logging.getLogger("genens")
            logger.debug(f"Timeouted:\n {workflow}")
            print(f"Timeouted:\n {workflow}")

            res = None

        return res
    return with_timeout


def eval_time(fn):
    @wraps(fn)
    def with_time(self, workflow, *args, **kwargs):
        start_time = time.process_time()

        res = fn(self, workflow, *args, **kwargs)
        if res is None or np.isnan(res):
            return None

        elapsed_time = time.process_time() - start_time

        logger = logging.getLogger("genens")
        logger.debug(f"Score: {res}, evaluation time - {elapsed_time}s,\n{workflow}")

        return res, np.log(elapsed_time)

    return with_time


def default_score(workflow, test_X, test_y):
    """
    Default scoring method for evaluators.

    :param workflow: Workflow to be evaluated.
    :param test_X: Features.
    :param test_y: Target vector.
    :return: Accuracy score of ``workflow`` on the data.
    """
    res_y = workflow.predict(test_X)
    return accuracy_score(test_y, res_y)


def _simple_eval(workflow, train_X, train_y, test_X, test_y, scorer=None):
    workflow.fit(train_X, train_y)
    if scorer is not None:
        return scorer(workflow, test_X, test_y)

    return default_score(workflow, test_X, test_y)


class EvaluatorBase(ABC):
    """
    Evaluator interface which is used in Genens during individual evaluation.
    """
    def __init__(self, timeout_s=None):
        self.train_X = None
        self.train_y = None

        self.timeout = timeout_s

    def __repr__(self):
        res = "{}".format(self.__class__.__name__)
        return res

    def fit(self, train_X, train_y):
        """
        Sets the training data for evaluation.

        :param train_X: Features.
        :param train_y: Target vector.
        """
        self.train_X = train_X
        self.train_y = train_y

    @abstractmethod
    def evaluate(self, workflow, scorer=None):
        """
        Evaluates a pipeline (workflow) learned on the training data from this evaluator. Evaluation method
        differs between different evaluators.
        :param workflow: Workflow to be evaluated.
        :param scorer: Scorer to be used, default is the accuracy scorer.
        :return: Score of the workflow.
        """
        pass

    @abstractmethod
    def reset(self):
        pass

    def check_is_fitted(self):
        if self.train_X is None or self.train_y is None:
            raise ValueError("Evaluator is not fitted with training data.")  # TODO specific

    @timeout
    @eval_time
    def score(self, workflow, scorer=None):
        """
        Computes the score of a pipeline (workflow).
        :param workflow: Workflow to be evaluated.
        :param scorer: Scorer to be used, default is the accuracy scorer.
        :return: Score of the workflow.
        """
        self.check_is_fitted()

        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')

                return self.evaluate(workflow, scorer=scorer)
        except Exception as e:
            logger = logging.getLogger("genens")
            logger.debug(f"Workflow failed:\n {e}\nWorkflow: {workflow}.")
            return None


class CrossValEvaluator(EvaluatorBase):
    """
    Evaluator which performs a k-fold cross-validation.
    """
    def __init__(self, cv_k=7, timeout_s=None):
        super().__init__(timeout_s)

        if cv_k < 0:
            raise AttributeError("Cross validation k must be greater than 0.")

        self.cv_k = cv_k

    def __repr__(self):
        res = super().__repr__()
        res += ", cv_k: {}".format(self.cv_k)
        return res

    def evaluate(self, workflow, scorer=None):
        scores = cross_val_score(workflow, self.train_X, self.train_y,
                                 cv=self.cv_k, scoring=scorer, error_score='raise')
        return np.mean(scores)

    def reset(self):
        pass


class FixedTrainTestEvaluator(EvaluatorBase):
    """
    Evaluator that uses fixed train and test samples.
    """
    def __init__(self, test_size=None, random_state=None, timeout_s=None):
        super().__init__(timeout_s)

        self.test_X = None
        self.test_y = None

        self.test_size = test_size
        self.random_state = random_state

    def __repr__(self):
        res = super().__repr__()
        res += ", test_size: {}".format(self.test_size)
        return res

    def fit(self, train_X, train_y):
        self.train_X, self.test_X, self.train_y, self.test_y = \
            train_test_split(train_X, train_y, test_size=self.test_size,
                             random_state=self.random_state)

    def check_is_fitted(self):
        super().check_is_fitted()
        if self.test_X is None or self.test_y is None:
            raise ValueError("Test data missing in evaluator.")

    def evaluate(self, workflow, scorer=None):
        return _simple_eval(workflow, self.train_X, self.train_y, self.test_X, self.test_y,
                            scorer=scorer)

    def reset(self):
        pass


class RandomTrainTestEvaluator(EvaluatorBase):
    """
    Evaluator which performs the train-test split for every evaluation.
    """
    def __init__(self, test_size=None, random_state=None, timeout_s=None):
        super().__init__(timeout_s)

        self.test_size = test_size

        self.rng = None
        if random_state is not None:
            if isinstance(random_state, np.random.RandomState):
                self.rng = random_state

            self.rng = np.random.RandomState(random_state)

    def __repr__(self):
        res = super().__repr__()
        res += ", test_size: {}".format(self.test_size)
        return res

    def evaluate(self, workflow, scorer=None):
        # random state is set only in the constructor
        train_X, test_X, train_y, test_y = train_test_split(self.train_X, self.train_y,
                                                            test_size=self.test_size,
                                                            random_state=self.rng)
        return _simple_eval(workflow, train_X, train_y, test_X, test_y,
                            scorer=scorer)

    def reset(self):
        pass


class TrainTestEvaluator(EvaluatorBase):
    """
    Evaluator with a specified validation set.
    """
    def __init__(self, test_X, test_y, timeout_s=None):
        super().__init__(timeout_s)

        # training set is provided in fit, validation set is set on initialization
        self.test_X = test_X
        self.test_y = test_y

    def evaluate(self, workflow, scorer=None):
        return _simple_eval(workflow, self.train_X, self.train_y, self.test_X, self.test_y,
                            scorer=scorer)

    def reset(self):
        pass


class DataSampler:
    """
    Provides random samples of a given dataset.
    """
    def __init__(self, sample_size=0.20, random_state=None, stratified=True):
        self.full_X = None
        self.full_y = None

        self.sample_size = sample_size
        self.stratified = stratified

        self.rng = None
        if random_state is not None:
            if isinstance(random_state, np.random.RandomState):
                self.rng = random_state

            self.rng = np.random.RandomState(random_state)

    def fit(self, full_X, full_y):
        self.full_X = full_X
        self.full_y = full_y

    def generate_sample(self):
        """
        Generates a random sample from full data.
        :return: A random sample.
        """
        stratify = self.full_y if self.stratified else None

        _, sample_X, _, sample_y = train_test_split(self.full_X, self.full_y,
                                                    test_size=self.sample_size,
                                                    random_state=self.rng,
                                                    stratify=stratify)
        return sample_X, sample_y


class SampleCrossValEvaluator(CrossValEvaluator):
    """
    Evaluator which uses a k-fold cross-validation based on sampling.
    """
    def __init__(self, cv_k=7, timeout_s=None, sample_size=0.20, per_gen=False, random_state=None):
        super().__init__(cv_k=cv_k, timeout_s=timeout_s)

        self.per_gen = per_gen
        self.sampler = DataSampler(sample_size=sample_size, random_state=random_state)

    def __repr__(self):
        res = super().__repr__()
        res += ", sample_size: {}".format(self.sampler.sample_size)
        res += ", per_gen: {}".format(self.per_gen)
        return res

    def _reset_data(self):
        self.train_X, self.train_y = self.sampler.generate_sample()

    def fit(self, train_X, train_y):
        self.sampler.fit(train_X, train_y)
        self._reset_data()

    def evaluate(self, workflow, scorer=None):
        # generate sample for every evaluation
        if not self.per_gen:
            self._reset_data()

        return super().evaluate(workflow, scorer=scorer)

    def reset(self):
        # generate sample once per reset
        if self.per_gen:
            self._reset_data()


class SampleTrainTestEvaluator(FixedTrainTestEvaluator):
    """
    Evaluator which uses a train-test evaluation on samples.
    """
    def __init__(self, test_size=None, timeout_s=None, sample_size=0.20,
                 per_gen=False, random_state=None):

        self.per_gen = per_gen
        self.sampler = DataSampler(sample_size=sample_size, random_state=random_state)

        # rng should be set in DataSampler ctor
        super().__init__(test_size=test_size, random_state=self.sampler.rng, timeout_s=timeout_s)

    def __repr__(self):
        res = super().__repr__()
        res += ", sample_size: {}".format(self.sampler.sample_size)
        res += ", per_gen: {}".format(self.per_gen)
        return res

    def _fit_sample(self):
        train_X, train_y = self.sampler.generate_sample()

        super().fit(train_X, train_y)

    def fit(self, train_X, train_y):
        self.sampler.fit(train_X, train_y)

        self._fit_sample()

    def evaluate(self, workflow, scorer=None):
        # one sample per evaluation
        if not self.per_gen:
            self._fit_sample()

        return super().evaluate(workflow, scorer=scorer)

    def reset(self):
        # generate new fixed sample
        if self.per_gen:
            self._fit_sample()


_eval_names = {
    'crossval': CrossValEvaluator,
    'fixed': FixedTrainTestEvaluator,
    'per_ind': RandomTrainTestEvaluator,
    'train_test': TrainTestEvaluator,
    'sample_crossval': SampleCrossValEvaluator,
    'sample_train_test': SampleTrainTestEvaluator
}


def get_evaluator_cls(name):
    return _eval_names[name]



import logging

from functools import wraps


def set_log_handler(func):
    @wraps(func)
    def with_set_log_handler(*args, **kwargs):
        log_setup = kwargs.pop('log_setup', None)
        if log_setup is None:
            return func(*args, **kwargs)

        handl = log_setup()
        res = func(*args, **kwargs)

        # remove unused handler
        logger = logging.getLogger("genens")
        logger.removeHandler(handl)
        return res

    return with_set_log_handler



# -*- coding: utf-8 -*-

import argparse

import inspect
import json

import itertools

import os

from sklearn.metrics import make_scorer

import pickle
import time

from genens import GenensClassifier, GenensRegressor
from genens.workflow.evaluate import get_evaluator_cls
from genens.config import clf_config
from genens.render.plot import export_plot
from genens.render.graph import create_graph
from genens.tests.load_datasets import load_dataset


def run_tests(estimators, train_X, train_y, out_dir, test_X=None, test_y=None):
    try:
        os.mkdir(out_dir)
    except FileExistsError:
        print("Test directory already exists.")
    except OSError as e:
        print("\nCannot create test directory.")
        raise e

    for i, (est, kwarg_dict) in enumerate(estimators):
        try:
            test_dir = out_dir + '/{}'.format(i)
            os.mkdir(test_dir)
        except FileExistsError:
            # skip finished tests
            print("\nTest {} skipped.".format(i))
            continue
        except OSError as e:
            print("\nCannot create directory for test {}".format(i))
            raise e
        else:
            print("\nRunning test {}".format(i))

        run_once(est, train_X, train_y, kwarg_dict, test_dir, test_X=test_X, test_y=test_y)


def evaluate_workflows(clf, features, target, scorer=None, method='crossval', **kwargs):
    wfs = clf.get_best_pipelines()[:5]

    cls = get_evaluator_cls(method)

    eval = cls(**kwargs)
    eval.fit(features, target)

    s = clf.scorer if scorer is None else scorer
    return [eval.score(wf, s) for wf in wfs]


def run_once(estimator, train_X, train_y, kwarg_dict, out_dir, test_X=None, test_y=None):
    # start test time measurement
    start_time = time.time()

    if test_X is not None and test_y is not None:
        estimator.setup_test_stats(train_X, train_y, test_X, test_y)

    estimator.fit(train_X, train_y)

    # end test time measurement
    elapsed_time = time.time() - start_time

    print('Evolution time: {}'.format(elapsed_time))

    # test the pipelines
    with open(out_dir + '/pipe-eval.txt', 'w+') as out_file:
        # default evaluation methods
        if test_X is not None and test_y is not None:
            res = evaluate_workflows(estimator, train_X, train_y, method='train_test',
                                     test_X=test_X, test_y=test_y)
        else:
            res = evaluate_workflows(estimator, train_X, train_y, method='crossval', cv_k=10)

        for i, val in enumerate(res):
            out_file.write("Pipeline {} score: {}\n".format(i, val))

    # write config to output dir
    with open(out_dir + '/config.txt', 'w+') as out_file:
        out_file.write(str(kwarg_dict) + '\n')

    # write logbook string representation to output dir
    with open(out_dir + '/logbook.txt', 'w+') as log_file:
        log_file.write('Test time: {}\n\n'.format(elapsed_time))
        log_file.write(estimator.logbook.__str__() + '\n')

    # pickle best pipelines and print to file
    with open(out_dir + '/pipelines.txt', 'w+') as out_file:
        for i, pipe in enumerate(estimator.get_best_pipelines()[:5]):
            with open(out_dir + '/pipeline{}.pickle'.format(i), 'wb') as pickle_file:
                pickle.dump(pipe, pickle_file, pickle.HIGHEST_PROTOCOL)

            # todo check with wilt
            # pprint.pprint(pipe, out_file)
            out_file.write(repr(pipe))
            out_file.write('\n\n')

    # evolution plot
    export_plot(estimator, out_dir + '/result.png')

    # individual fitness values
    with open(out_dir + '/ind-fitness.txt', 'w+') as out_file:
        for i, ind in enumerate(estimator.get_best_pipelines(as_individuals=True)[:5]):
            out_file.write('Individual {}: Score {}, Test {}\n'.format(i, ind.fitness.values,
                                                                       ind.test_stats))
            create_graph(ind, out_dir + '/graph{}.png'.format(i))


def create_scorer(scorer_path, scorer_kwargs):
    tmp_path = scorer_path.split('.')
    op_str = tmp_path.pop()
    import_str = '.'.join(tmp_path)

    try:
        exec('from {} import {}'.format(import_str, op_str))
    except ImportError as e:
        print("Could not import scoring function.")
        raise e

    cls = eval(op_str)

    return make_scorer(cls, **scorer_kwargs)


def load_config(cmd_args):
    # read config
    with open(cmd_args.file) as f:
        config = json.load(f)

    params = config['parameters']

    # set up scorer
    if 'scorer' in config.keys():
        func = config['scorer']['func']
        scorer_args = config['scorer']['kwargs']

        scorer = create_scorer(func, scorer_args)
        params['scorer'] = scorer

    def product_dict(**kwargs):
        """
        Makes a dictionary product; computes the product of all lists in the
        arguments, for every element of the product returns a dictionary of these
        arguments. If an argument isn't a list, it is added to every element
        of the product.
        :param kwargs: Keyword argument dictionary.
        :return: Keyword argument product.
        """
        only_lists = {key: val for key, val in kwargs.items() if isinstance(val, list)}
        other_args = {key: val for key, val in kwargs.items() if key not in only_lists.keys()}

        keys = only_lists.keys()
        for values in itertools.product(*only_lists.values()):
            yield dict(zip(keys, values), **other_args)

    def obj_kwargs_product(cls, kwargs_dict):
        for kwargs in product_dict(**kwargs_dict):
            yield cls(**kwargs)

    if 'evaluator' in config.keys():
        eval_cls = get_evaluator_cls(config['evaluator']['func'])
        params['evaluator'] = list(obj_kwargs_product(eval_cls, config['evaluator']['kwargs']))

    if 'group_weights' in config.keys():
        params['config'] = [clf_config(gweight) for gweight in
                            product_dict(**config['group_weights'])]

    param_product = product_dict(**params)

    def clf_iterate(param_prod):
        for kwargs in param_prod:
            if cmd_args.regression:
                yield GenensRegressor(**kwargs), {**kwargs}
            else:
                yield GenensClassifier(**kwargs), {**kwargs}

    datasets = config['datasets']

    for dataset in datasets:
        if 'split_validation' in dataset and dataset['split_validation']:
            train_X, train_Y, test_X, test_Y = load_dataset(**dataset)

            run_tests(clf_iterate(param_product), train_X, train_Y,
                      cmd_args.out + '/' + dataset['dataset_name'],
                      test_X=test_X, test_y=test_Y)
        else:
            features, target = load_dataset(**dataset)
            run_tests(clf_iterate(param_product), features, target,
                      cmd_args.out + '/' + dataset['dataset_name'])


def load_from_args(cmd_args):
    # TODO rewrite if needed (only for 'run' case)

    arg_dict = vars(cmd_args)

    if cmd_args.regression:
        cls = GenensRegressor
    else:
        cls = GenensClassifier

    cls_params = inspect.signature(cls.__init__).parameters
    kwargs = dict((key, val) for key, val in arg_dict.items() if key in cls_params)

    dataset = cmd_args.dataset
    train_X, train_Y, test_X, test_Y = load_dataset(dataset, split_validation=True)

    run_once(cls(**kwargs), train_X, train_Y, kwargs, cmd_args.out,
             test_X=test_X, test_y=test_Y)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Run Genens on datasets.")
    parser.add_argument('--out', required=True)
    parser.add_argument('--regression', action='store_true')

    subparsers = parser.add_subparsers()

    # pass arguments in a config file
    parser_config = subparsers.add_parser('config')
    parser_config.add_argument('file', type=str, help='Configuration file location.')

    parser_config.set_defaults(func=load_config)

    parser_run = subparsers.add_parser('run')

    # pass arguments on command line
    parser_run.add_argument('--dataset', type=str, help='Dataset name.', required=True)
    parser_run.add_argument('--n_jobs', type=int, default=1, help='Sets n_jobs of Genens.')
    parser_run.add_argument('--cx_pb', type=float, default=0.5, help='Sets crossover probability.')
    parser_run.add_argument('--mut_pb', type=float, default=0.1, help='Sets mutation probability.')

    parser_run.add_argument('--mut_args_pb', type=float, default=0.3,
                            help='Sets argument mutation probability.')

    parser_run.add_argument('--scorer', type=str, help='Sets a custom scorer.')
    parser_run.add_argument('--pop_size', type=int, default=100, help='Sets population size.')

    parser_run.add_argument('--n_gen', type=int, default=10,
                            help='Sets the number of generations.')

    parser_run.add_argument('--hc_repeat', type=int, default=0,
                            help='Sets the number of hill-climbing repetitions.')

    parser_run.add_argument('--hc_keep_last', action='store_true',
                            help="If true, mutates the individual if hill-climbing hasn't been successful.")

    parser_run.set_defaults(func=load_from_args)

    args = parser.parse_args()
    args.func(args)



import numpy as np
import random

from genens.gp.types import DeapTreeIndividual


def _choose_group_weighted(config, groups):
    # weighted choice
    total_sum = np.sum((config.group_weights[group] for group in groups))
    rand_val = random.random() * total_sum

    # determine which group was chosen
    group_chosen = None
    partial_sum = 0.0
    for group in groups:
        partial_sum += config.group_weights[group]
        if partial_sum > rand_val:
            return group

    if group_chosen is None:
        raise RuntimeError("Invalid weight sum.")  # should never get here


def choose_prim(config, prim_list, weighted=True, use_groups=True):
    """
    Performs a selection from list of primitives.

    :param GenensConfig config: Configuration of Genens which contains the weight configuration.
    :param list prim_list: List of primitives to choose from.
    :param bool weighted: Determines whether the selection is weighted (according to primitive group weights).
    :param bool use_groups: If false, primitive groups are ignored
    :return GpPrimitive: The selected primitive.
    """
    if not use_groups:
        return random.choice(prim_list)

    group_names = list({prim.group for prim in prim_list})  # primitive groups to choose from

    if weighted:
        group_chosen = _choose_group_weighted(config, group_names)
    else:
        group_chosen = random.choice(group_names)

    prim_possible = [prim for prim in prim_list if prim.group == group_chosen]
    return random.choice(prim_possible)


def gen_tree(config, max_height=None, max_arity=None, first_type='out', weighted=True, use_groups=True):
    """
    Creates a random tree individual to be used in the DEAP framework.

    :param config: Configuration of nodes, keyword arguments and arity and height limits.
    :param max_height: Height limit of the tree; if not specified, the configuration limit is used.
    :param max_arity: Arity limit of function nodes; if not specified, the configuration limit is used.
    :param first_type: Output type of the root of the tree.
    :param weighted: Determines whether the node selection is weighted.
    :param bool use_groups: If false, primitive groups are ignored
    :return: A random tree individual.
    """
    tree_list = []
    type_stack = [(first_type, 1, 1)]
    tree_height = 0

    if max_height is None:
        max_height = config.max_height

    if max_arity is None:
        max_arity = config.max_arity

    while len(type_stack):
        next_type, ar, h = type_stack.pop()
        node_depth = h - 1
        tree_height = max(h, tree_height)

        # more children should be generated
        if ar - 1 > 0:
            type_stack.append((next_type, ar - 1, h))

        # choose only terminals in the last level
        if h < max_height - 1:
            choose_from = config.full_config[next_type]
        else:
            choose_from = config.term_config[next_type]

        # template of the next primitive
        next_prim_t = choose_prim(config, choose_from, weighted=weighted, use_groups=use_groups)
        prim = next_prim_t.create_primitive(node_depth, max_arity,
                                            config.kwargs_config[next_prim_t.name])

        # append child types and arities to the stack
        if prim.arity > 0:
            for child_type in prim.node_type[0]:
                type_stack.append((child_type.name, child_type.arity, h + 1))

        tree_list.append(prim)

    return DeapTreeIndividual(list(reversed(tree_list)), tree_height)


def swap_subtrees(tree_1, tree_2, ind_1, ind_2, keep_2=True):
    """
    Swaps subtrees of argument trees. Subtree position is determined by
    ``ind_1`` and ``ind_2``.

    If ``keep_2`` is False, the subtree from
    ``tree_1`` is not inserted into ``tree_2`` and only the first tree
    is returned.

    :param GpTreeIndividual tree_1: First tree.
    :param GpTreeIndividual tree_2: Second tree.
    :param int ind_1: Index of the subtree of the first subtree.
    :param int ind_2: Index of the subtree of the second tree.
    :param bool keep_2: Indicates whether the second modified tree would be returned.
    :return (GpTreeIndividual, GpTreeIndividual) or (GpTreeIndividual,):
        Returns both trees with swapped subtrees or only the first tree with
        a subtree inserted from ``tree_2`` (according to ``keep_2``).
    """

    # update node heights
    root_height_1 = tree_1.primitives[ind_1].depth
    root_height_2 = tree_2.primitives[ind_2].depth
    height_diff = root_height_1 - root_height_2

    def move_node(prim, diff):
        prim.height = prim.depth + diff
        return prim

    # subtree indices
    ind_begin_1, _ = tree_1.subtree(ind_1)
    ind_begin_2, _ = tree_2.subtree(ind_2)

    ind_end_1 = ind_1 + 1
    ind_end_2 = ind_2 + 1

    # insert into tree_1 - copy subtree from tree_2
    subtree_2 = [move_node(prim, height_diff)
                 for prim in tree_2.primitives[ind_begin_2 : ind_end_2]]

    # insert into tree_2
    if keep_2:
        subtree_1 = (move_node(prim, -height_diff)
                     for prim in tree_1.primitives[ind_begin_1 : ind_end_1])

        tree_2.primitives[ind_begin_2 : ind_end_2] = subtree_1
        tree_2.max_height = max(prim.depth + 1 for prim in tree_2.primitives)  # update height

    # insert into tree_1 - insert subtree
    tree_1.primitives[ind_begin_1: ind_end_1] = subtree_2
    tree_1.max_height = max(prim.depth + 1 for prim in tree_1.primitives)  # update height

    return tree_1, tree_2 if keep_2 else tree_1,



# -*- coding: utf-8 -*-
from sklearn.linear_model import LinearRegression, Ridge, Lasso, ElasticNet, Lars

from ..config.utils import get_default_config, predictor_primitive, predictor_terminal
from ..config.utils import estimator_func
from ..config.utils import ensemble_func
from ..config.utils import ensemble_primitive
from ..config.utils import transformer_primitive

from sklearn import decomposition
from sklearn import feature_selection
from sklearn import preprocessing

from sklearn import ensemble

import warnings

from ..gp.types import GpFunctionTemplate, TypeArity


def regr_config():
    """
    Creates Genens configuration for regression problems.
    It does not contain regressors, will be extended in future releases.

    :return GenensConfig: Genens configuration for regression problems.
    """

    warnings.warn("This configuration is incomplete. You can extend it with regressors to use it; in next releases,"
                  " it will be finished.")

    config = get_default_config()

    ensembles_func = {
        'voting': ensemble_func(ensemble.VotingRegressor),
        'stacking': ensemble_func(ensemble.StackingRegressor)
    }

    # TODO
    regr_func = {
        'ada': estimator_func(ensemble.AdaBoostRegressor),
        'bagging': estimator_func(ensemble.BaggingRegressor),
        'LinearRegression': estimator_func(LinearRegression),
        'Ridge': estimator_func(Ridge),
        'Lasso': estimator_func(Lasso),
        'ElasticNet': estimator_func(ElasticNet),
        'Lars': estimator_func(Lars)

    }

    transform_func = {
                "NMF": estimator_func(decomposition.NMF),
                "FA": estimator_func(decomposition.FactorAnalysis),
                "FastICA": estimator_func(decomposition.FastICA),
                "PCA": estimator_func(decomposition.PCA),
                #"KernelPCA" : estimator_func(decomposition.KernelPCA),
                "kBest": estimator_func(feature_selection.SelectKBest),
                "MaxAbsScaler": estimator_func(preprocessing.MaxAbsScaler),
                "MinMaxScaler":estimator_func(preprocessing.MinMaxScaler),
                "Normalizer": estimator_func(preprocessing.Normalizer),
                "StandardScaler": estimator_func(preprocessing.StandardScaler)
    }

    ensemble_kwargs = {
        'voting': {
            # 'soft' not included - a lot of classifiers does not support predict_proba
            'voting': ['hard', 'soft']
        },
        'stacking': {
            'stack_method': ['auto', 'predict_proba', 'predict'],
            'cv': [3, 5]
        }
    }

    regr_kwargs = {
        'ada': {
            'n_estimators': [5, 10, 50, 100, 200],
            'loss': ['linear', 'square', 'exponential']
        },
        'bagging': {
            'n_estimators': [5, 10, 50, 100, 200]
        }
    }

    transform_kwargs = {
        'NMF': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'solver': ['cd', 'mu']
        },
        'FA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
        },
        'FastICA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],

        },
        'PCA': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'whiten': [False, True],
        },
        # 'KernelPCA': {
        #   'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
        #   'kernel': ['linear', 'poly', 'rbf', 'sigmoid', 'cosine']
        # },
        'kBest': {
            'feat_frac': [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            'score_func': [feature_selection.chi2, feature_selection.f_classif]
        },
        'MaxAbsScaler': {},
        'MinMaxScaler': {},
        "Normalizer": {},
        "StandardScaler": {}
    }

    func_dict = {**ensembles_func, **regr_func, **transform_func}
    kwargs_dict = {**ensemble_kwargs, **regr_kwargs, **transform_kwargs}

    config.add_functions_args(func_dict, kwargs_dict)

    # ensemble config
    config.add_primitive(ensemble_primitive('voting', (2, 'n')))
    config.add_primitive(
        GpFunctionTemplate('stacking',
                           [TypeArity('out', (1, 'n')), TypeArity('out', 1)],
                           'ens',
                           group='ensemble')
    )

    # regressor config
    config.add_primitive(predictor_primitive('ada', 1))
    config.add_primitive(predictor_primitive('bagging', 1))

    config.add_primitive(predictor_primitive('ada', 1), term_only=True)
    config.add_primitive(predictor_primitive('bagging', 1), term_only=True)

    # transformer config
    config.add_primitive(transformer_primitive("NMF", 'featsel'))
    config.add_primitive(transformer_primitive("FA", 'featsel'))
    config.add_primitive(transformer_primitive("FastICA", 'featsel'))
    config.add_primitive(transformer_primitive("PCA", 'featsel'))
    #config.add_primitive(transformer_primitive("KernelPCA", 'featsel'))
    config.add_primitive(transformer_primitive("kBest", 'featsel'))

    config.add_primitive(transformer_primitive("MaxAbsScaler", 'scale'))
    config.add_primitive(transformer_primitive("MinMaxScaler", 'scale'))
    config.add_primitive(transformer_primitive("Normalizer", 'scale'))
    config.add_primitive(transformer_primitive("StandardScaler", 'scale'))

    return config



# -*- coding: utf-8 -*-

"""
This module contains the Genens estimators.
"""

from .base import GenensBase
from .config import clf_config
from .config import regr_config

from sklearn.base import ClassifierMixin, RegressorMixin


class GenensClassifier(GenensBase, ClassifierMixin):
    def __init__(self, config=None, n_jobs=1, cx_pb=0.5, mut_pb=0.3, mut_args_pb=0.9, mut_node_pb=0.9,
                 scorer=None, pop_size=100, n_gen=15, hc_repeat=0, hc_keep_last=False,
                 mut_multiple_args=False, mut_multiple_nodes=False,
                 weighted=True, use_groups=True,
                 max_height=None, max_arity=None, timeout=None, evaluator=None, hc_n_nodes=3, hc_mut_pb=0.2,
                 max_evo_seconds=None, log_path=None,
                 **kwargs):
        if config is None:
            config = clf_config()

        super().__init__(config,
                         n_jobs=n_jobs,
                         cx_pb=cx_pb,
                         mut_pb=mut_pb,
                         mut_args_pb=mut_args_pb,
                         mut_node_pb=mut_node_pb,
                         scorer=scorer,
                         pop_size=pop_size,
                         n_gen=n_gen,
                         hc_repeat=hc_repeat,
                         hc_keep_last=hc_keep_last,
                         hc_n_nodes=hc_n_nodes,
                         hc_mut_pb=hc_mut_pb,
                         mut_multiple_args=mut_multiple_args,
                         mut_multiple_nodes=mut_multiple_nodes,
                         weighted=weighted,
                         use_groups=use_groups,
                         max_height=max_height,
                         max_arity=max_arity,
                         timeout=timeout,
                         evaluator=evaluator,
                         max_evo_seconds=max_evo_seconds,
                         log_path=log_path,
                         **kwargs)


class GenensRegressor(GenensBase, RegressorMixin):
    def __init__(self, config=None, n_jobs=1, cx_pb=0.5, mut_pb=0.3, mut_args_pb=0.9, mut_node_pb=0.9,
                 scorer=None, pop_size=100, n_gen=15, hc_repeat=0, hc_keep_last=False,
                 mut_multiple_args=False, mut_multiple_nodes=False,
                 weighted=True, use_groups=True,
                 max_height=None, max_arity=None, timeout=None, evaluator=None,
                 hc_n_nodes=3, hc_mut_pb=0.1, max_evo_seconds=None, **kwargs):
        if config is None:
            config = regr_config()

        super().__init__(config,
                         n_jobs=n_jobs,
                         cx_pb=cx_pb,
                         mut_pb=mut_pb,
                         mut_args_pb=mut_args_pb,
                         mut_node_pb=mut_node_pb,
                         scorer=scorer,
                         pop_size=pop_size,
                         n_gen=n_gen,
                         hc_repeat=hc_repeat,
                         hc_keep_last=hc_keep_last,
                         hc_n_nodes=hc_n_nodes,
                         hc_mut_pb=hc_mut_pb,
                         mut_multiple_args=mut_multiple_args,
                         mut_multiple_nodes=mut_multiple_nodes,
                         weighted=weighted,
                         use_groups=use_groups,
                         max_height=max_height,
                         max_arity=max_arity,
                         timeout=timeout,
                         evaluator=evaluator,
                         max_evo_seconds=max_evo_seconds,
                         **kwargs)



# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import openml
import os
import pandas as pd
import seaborn as sns


def get_score_stats(columns, dirs):

    df = pd.DataFrame()
    for col_name, dir_path in zip(columns, dirs):
        file_paths = os.listdir(dir_path)
        file_paths = ["/".join([dir_path, f]) for f in file_paths]

        df[col_name] = read_score_list(file_paths)

    return df


def read_score_list(file_list):
    score_list = []
    for file_name in file_list:
        max_val = 0.0
        with open(file_name, 'r') as in_file:
            for line in in_file:
                val = float(line)
                max_val = max(val, max_val)

        score_list.append(max_val)
    return score_list


def boxplot_get_stats(df):
    res = []
    for column in df.columns:
        bx = plt.boxplot(df[column], notch=True)

        c_min = np.min(df[column])
        c_max = np.max(df[column])

        bx_data = bx['boxes'][0].get_ydata()
        c_cilo = bx_data[2]
        c_med = bx_data[3]
        c_cihi = bx_data[4]

        res.append([c_min, c_cilo, c_med, c_cihi, c_max])

    res_df = pd.DataFrame(data=res, columns=['minimum', 'ci_low', 'median', 'ci_high', 'maximum'], index=df.columns)
    return res_df.round(4)


def boxplot_compare_columns(df, out_path, out_name='outbox.png', x_axis='Evaluation strategy',
                            eval_metric='Predictive accuracy'):
    sns.set()

    bx_plot = sns.boxplot(x='variable', y='value', data=pd.melt(df), notch=True)
    plt.xlabel(x_axis)
    plt.ylabel(eval_metric)

    plt.tight_layout()
    plt.savefig('/'.join([out_path, out_name]))


def get_openml_stats():

    benchmark_suite = openml.study.get_study('OpenML-CC18', 'tasks')

    accuracy_df = pd.DataFrame(columns=['task_id', 'name', 'accuracy'])

    for task_id in benchmark_suite.tasks:
        print('Task {}'.format(task_id))

        evaluations = openml.evaluations.list_evaluations('predictive_accuracy', task=[task_id])

        vals = [(e.task_id, e.data_name, e.value) for e in evaluations.values()]
        df = pd.DataFrame(vals, columns=['task_id', 'name', 'accuracy'])

        accuracy_df = accuracy_df.append(df)

    return accuracy_df


if __name__ == "__main__":
    df = get_openml_stats()
    df.to_csv("openml-data.csv", sep=';')



# -*- coding: utf-8 -*-
"""
Module for visualization of the trees.
"""

import pygraphviz as pgv

from genens.gp.types import GpTreeIndividual, GpPrimitive


def create_graph(gp_tree, file_name):
    graph = pgv.AGraph()
    i = 0

    def add_node(node, child_list):
        nonlocal i
        node_id = i
        i = i + 1

        graph.add_node(node_id, label=node.name)

        for val in child_list:
            graph.add_edge(node_id, val)

        return node_id

    gp_tree.run_tree(add_node)

    graph.layout(prog='dot')
    graph.draw(file_name, prog='dot')
    graph.close()


def tree_str(gp_tree: GpTreeIndividual, with_hyperparams=False):

    def get_tree(node, ch_l):
        return node, ch_l

    def print_node(node: GpPrimitive, ch_l, indent, str_res):
        str_res += "  " * indent + node.name + "\n"
        if with_hyperparams:
            for k, val in node.obj_kwargs.items():
                str_res += "  " * (indent + 1) + "| " + f"{k}: {val}\n"

        for child, next_list in ch_l:
            str_res = print_node(child, next_list, indent + 1, str_res)

        return str_res

    root, child_list = gp_tree.run_tree(get_tree)
    return print_node(root, child_list, 0, "")



# -*- coding: utf-8 -*-

"""
This module contains helper functions for the creation of GP primitives and
construction functions for most common methods used in scikit-learn.

The functions create node templates and wrappers of ensembles, simple predictors and transformers.
Node templates are used in the GP evolution process, wrapper functions are used to convert nodes
to machine learning workflow methods.

Node templates are specified by input types, arities (one arity/arity range per type) and output types.

Wrapper functions have the signature ``func(child_list, kwarg_dict)``, where ``child_list`` are results
of other wrapper functions applied on child nodes and ``kwarg_dict`` is a dictionary of evolved keyword
arguments.
"""

from functools import partial

from ..gp.types import GpFunctionTemplate, GpTerminalTemplate, TypeArity
from ..workflow.model_creation import create_pipeline, create_stacking
from ..workflow.model_creation import create_data_union
from ..workflow.model_creation import create_transform_list
from ..workflow.model_creation import create_empty_data
from ..workflow.model_creation import create_estimator
from ..workflow.model_creation import create_ensemble

from warnings import warn


def default_group_weights(*args, groups=None):
    """
    Verifies whether all GpPrimitive instances contained in ``args`` (dictionaries of lists) belong to
    a particular group in ``groups``. If a GpPrimitive does not have a group specified, creates a new
    group for it with default weight.

    :param args: Dictionaries of lists of GpPrimitive (grouped by output types as in GenensConfig)
    :param groups: Dictionary of groups with weights
    :return: Adjusted groups
    """
    checked_groups = groups if groups is not None else {}
    for prim_list in args:
        # primitives are grouped by output nodes
        for out_prim_list in prim_list.values():
            for prim in out_prim_list:

                if prim.group is None:
                    # warn only if user provided config with missing values
                    if groups is not None:
                        warn("Primitive {} has no group, using its name as the group.".format(prim.name))

                    prim.group = prim.name

                if prim.group not in checked_groups.keys():
                    # warn only if user provided config with missing values
                    if groups is not None:
                        warn("Group {} is not present in the group list, adding with default"
                             "weight (1.0).".format(prim.group))

                    checked_groups[prim.group] = 1.0

    return checked_groups


class GenensConfig:
    """
    Configuration of Genens estimators. Contains settings of GP and configuration of
    methods which decode nodes into scikit-learn methods.
    """
    def __init__(self, func, full, term, kwargs_config, min_height=1, max_height=4, min_arity=2, max_arity=3,
                 group_weights=None):
        """
        Creates a new instance of a Genens configuration. If ``group_weights`` are specified,
        validity check is performed.

        :param dict func: Function configuration dictionary, keys correspond do a particular primitive name.
        :param dict full: Node configuration dictionary, nodes are grouped by output types,
                          contains node templates used during the grow phase.

        :param dict term: Node configuration dictionary, nodes are grouped by output types,
                          contains node templates used to terminate a tree (in the last height level).

        :param kwargs_config: Keyword arguments for nodes, which are passed to functions during the decoding.
                              For every argument, there is a list of possibles values to choose from during the
                              evolution.

        :param min_height: Minimum height of trees.
        :param max_height: Maximum height of trees.
        :param min_arity: Minimum arity of a function node.
        :param max_arity: Maximum arity of a function node.
        :param group_weights: Group weight configuration.
        """
        self.func_config = func

        self.full_config = full
        self.term_config = term

        self.kwargs_config = kwargs_config

        self.min_height = min_height
        self.max_height = max_height
        self.min_arity = min_arity
        self.max_arity = max_arity

        # perform check if specified
        if group_weights is None:
            self._group_weights = None
        else:
            self.group_weights = group_weights

    def __repr__(self):
        res = "max_height: {}, max_arity: {}".format(self.max_height, self.max_arity)
        res += ", group_weights: {}".format(str(self._group_weights))
        return res

    @property
    def group_weights(self):
        return self._group_weights

    @group_weights.setter
    def group_weights(self, value):
        """
        Sets the group weight and performs validity check, possibly adds a new group for primitives
        without a group specified.

        :param dict value: Dictionary of weights, keys are group names.
        """
        self._group_weights = default_group_weights(self.full_config, self.term_config,
                                                    groups=value)

    @group_weights.deleter
    def group_weights(self):
        del self._group_weights

    def add_primitive(self, prim, term_only=False):
        """
        Adds a new primitive to the configuration. If ``term_only`` is True,
        it is added only to the terminal set. If the primitive is a GpFunctionTemplate,
        it is added only to the grow set.

        :param GpPrimitive prim: Primitive to be added.
        :param bool term_only: Specifies whether the primitive should be added only to the terminal set.
        """
        if term_only and not isinstance(prim, GpTerminalTemplate):
            warn("Primitive is added neither to function nor terminal set.")

        if not term_only:
            out_list = self.full_config.setdefault(prim.out_type, [])
            out_list.append(prim)

        if isinstance(prim, GpTerminalTemplate):
            out_list = self.term_config.setdefault(prim.out_type, [])
            out_list.append(prim)

    def add_functions_args(self, func_dict, kwarg_dict):
        """
        Adds functions that decode a node along with keyword arguments for the node.
        The function has the signature ``func(child_list, kwargs)``, where child_list contains
        decoded values of child nodes, and kwargs are
        specific keyword arguments set during the evolution.

        :param func_dict: Dictionary of functions, the keys are names which should correspond to a node.
                          Every function must have a keyword argument dictionary specified.

        :param kwarg_dict: Dictionary of keyword argument dictionaries, the keys (names) corresponds to nodes
                           from grow/terminals sets and also to methods in the ``func_dict``.
        """
        for key, val in func_dict.items():
            if key in self.func_config.keys():
                raise ValueError("Cannot insert to func - duplicate value.")  # TODO specific

            kwarg_val = kwarg_dict.get(key, None)
            if kwarg_val is None:
                raise ValueError("Must provide keyword arguments for all names.")  # TODO specific

            self.func_config[key] = val
            self.kwargs_config[key] = kwarg_val


def get_default_config(group_weights=None):
    """
    Creates the default config with pipeline construction nodes and functions.
    
    :param group_weights: Group weights of the config.
    :return: The default configuration.
    """
    
    func_config = {
        'cPred': create_pipeline,
        'cPipe': create_pipeline,
        # 'dUnion': create_data_union,
        'cData': create_transform_list,
        'cFeatSelect': create_transform_list,
        'cScale': create_transform_list,
        'dTerm': create_empty_data
    }

    kwargs_config = {
        'cPred': {},
        'cPipe': {},
        'cData': {},
        'cFeatSelect': {},
        'cScale': {},
        # 'dUnion': {},
        'dTerm': {}
    }

    full_config = {
        'out': [
            GpFunctionTemplate('cPred', [TypeArity('ens', 1)], 'out',
                               group='pipeline'),
            GpFunctionTemplate('cPipe', [TypeArity('ens', 1), TypeArity('data', 1)], 'out',
                               group='pipeline'),
        ],
        'data': [
            # GpFunctionTemplate('dUnion', [TypeArity('data', (2,3))], 'data', group='union'),
            GpFunctionTemplate('cData', [TypeArity('featsel', 1), TypeArity('scale', 1)], 'data',
                               group='prepro'),
            GpFunctionTemplate('cFeatSelect', [TypeArity('featsel', 1)], 'data',
                               group='prepro'),
            GpFunctionTemplate('cScale', [TypeArity('scale', 1)], 'data',
                               group='prepro')
        ],
        'ens': []
    }

    term_config = {
        'out': [],
        'data': [],
        'ens': []
    }

    return GenensConfig(func_config, full_config, term_config, kwargs_config,
                        group_weights=group_weights)


def estimator_func(est_cls, **kwargs):
    """
    Creates a wrapper function which returns an instance of the argument estimator class.

    The function signature is ``func(child_list, kwarg_dict)``,
    keyword arguments in ``kwargs`` and ``kwarg_dict`` must be distinct.

    The ``child_list`` argument of the resulting function must be empty,
    as simple estimators cannot have sub-estimators.

    :param est_cls: Estimator class.
    :param kwargs: Keyword arguments of the estimator.
    :return: Function which constructs a new instance of the estimator.
    """
    return partial(create_estimator, est_cls, kwargs)


def ensemble_func(ens_cls, **kwargs):
    """
    Creates a wrapper function which returns an instance of the argument ensemble class.

    The function signature is ``func(child_list, kwarg_dict)``,
    keyword arguments in ``kwargs`` and ``kwarg_dict`` must be distinct.

    The ``child_list`` argument contains estimators (or pipelines) which will be
    set as the ``base_estimator`` or ``estimators`` of the ensemble.

    :param ens_cls: Ensemble class.
    :param kwargs: Keyword arguments of the ensemble.
    :return: Function which constructs a new instance of the ensemble.
    """
    return partial(create_ensemble, ens_cls, kwargs)


def stacking_func(ens_cls, **kwargs):
    return partial(create_stacking, ens_cls, kwargs)


def ensemble_primitive(ens_name, in_arity, in_type='out', out_type='ens', group='ensemble'):
    """
    Creates a function template which represents an ensemble. The template can be used
    to create GP primitives with variable arity and different keyword argument dictionaries.

    The node must have child nodes of a type ``in_type``, their count is specified
    by ``in_arity``. The ``out_type`` is 'ens' by default, which is the output type
    of predictors.

    :param group:
    :param str ens_name: Name of the ensemble.
    :param int, (int, int), (int, 'n') in_arity:
        Arity of input nodes, either a constant, or a range (inclusive), or a range
        without a specified upper bound (in the case of (int, 'n')).

    :param str in_type: Type of all child nodes.
    :param str out_type: Node output type.
    :return: Function template which can be used to create an ensemble primitive.
    """
    return GpFunctionTemplate(ens_name, [TypeArity(in_type, in_arity)], out_type,
                              group=group)


def predictor_primitive(p_name, group='predictor'):
    """
    Creates a terminal template which represents a simple predictor. The template
    can be used to create GP primitives with different keyword argument dictionaries.

    The node has the output type 'ens', which is the output type of predictors.

    :param group:
    :param str p_name: Name of the predictor.
    :return: Terminal template which can be used to create a predictor primitive.
    """
    return GpTerminalTemplate(p_name, 'ens', group=group)


def predictor_terminal(p_name, group='predictor'):
    """
    Creates a terminal template which represents a simple predictor. The template
    can be used to create GP primitives with different keyword argument dictionaries.

    The node has the output type 'out', which is the output type of pipelines.
    This node should be used when maximum tree height would be exceeded; in other
    cases, nodes returned by ``predictor_primitive`` should be used.

    :param group:
    :param str p_name: Name of the predictor.
    :return: Terminal template which can be used to create a predictor primitive.
    """
    return GpTerminalTemplate(p_name, 'out', group=group)


def transformer_primitive(t_name, out_type, group='transform'):
    """
    Creates a terminal template which represents a simple transformer. The template
    can be used to create GP primitives with different keyword argument dictionaries.

    The node has the output type 'data', which is the output type of transformer nodes.

    :param group:
    :param str t_name: Name of the transformer.
    :param str out_type:
        Name of the transformer output type (most common are 'featsel' for feature
        selectors and 'scale' for scaling transformers.
    :return: Terminal template which can be used to create a transformer primitive.
    """
    return GpTerminalTemplate(t_name, out_type, group=group)



import logging
import random
from genens.gp.tree import swap_subtrees


def crossover_one_point(gp_tree_1, gp_tree_2, min_node_depth=0):
    """
    Performs a crossover of two tree individuals.

    :param DeapTreeIndividual gp_tree_1: First individual.
    :param DeapTreeIndividual gp_tree_2: Second individual.
    :param int min_node_depth: Minimal crossover point depth in both trees.
    :return: Tree individuals modified by the crossover.
    """
    type_set_1 = {gp_ind.out_type for gp_ind in gp_tree_1.primitives if gp_ind.depth >= min_node_depth}
    type_set_2 = {gp_ind.out_type for gp_ind in gp_tree_2.primitives if gp_ind.depth >= min_node_depth}

    common_types = list(type_set_1.intersection(type_set_2))
    if not len(common_types):
        logger = logging.getLogger("genens")
        logger.debug(f"Crossover - no nodes of the same type:\n{gp_tree_1}\nx\n{gp_tree_2}")
        return gp_tree_1, gp_tree_2

    cx_type = random.choice(common_types)

    # nodes which can be roots of subtrees to be swapped
    eligible_1 = [ind for ind, node in enumerate(gp_tree_1.primitives)
                  if node.out_type == cx_type and node.depth >= min_node_depth]
    eligible_2 = [ind for ind, node in enumerate(gp_tree_2.primitives)
                  if node.out_type == cx_type and node.depth >= min_node_depth]

    cx_1 = random.choice(eligible_1)
    cx_2 = random.choice(eligible_2)

    return swap_subtrees(gp_tree_1, gp_tree_2, cx_1, cx_2)



# -*- coding: utf-8 -*-

"""
This module defines the structure of GP primitives.

The GP primitives are nodes with typed edges (parent input and child output types must match) and variable
arity (for a given type, its final arity can be chosen during the evolution process).

A ``GpPrimitive`` is a node whose types, arities and keyword arguments have been decided. To create such primitives,
it is possible to take use of templates. These contain possible values of arities, types and keyword arguments and
methods for choosing final values.

The primitive templates defined in this file are
    1) functions - inner nodes of the tree, transform input into output
    2) terminals - leaves of the tree, provide constant output.
"""


import random
import functools
from copy import deepcopy

from deap import base


class GpTreeIndividual:
    """Represents a tree individual used in the GP.
    The individual is a tree encoded as a list of ``GpPrimitive`` nodes.

    The list is a post-order representation of the tree. The tree can be
    uniquely reconstructed using the arity (and types) of primitives.
    """
    def __init__(self, prim_list, max_height):
        """
        Construct a GP tree from a list of primitives.

        :param list of GpPrimitive prim_list: Post-order representation of the tree.
        """
        self.primitives = prim_list
        self.max_height = max_height

        self.validate_tree()

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))

        return new

    def __eq__(self, other):
        if not isinstance(other, GpTreeIndividual):
            return False

        if self.primitives != other.primitives:
            return False

        return True

    def __repr__(self):
        return 'GpTreePrimitive(height={}, '.format(self.max_height)\
               + 'primitives=' + self.primitives.__repr__() + ')'

    def run_tree(self, node_func):
        """
        Applies a function with the signature ``func(node, child_list)`` on all
        nodes in the tree. The tree is traversed in post-order.

        The arguments of the function are a node and list of result values of its child nodes.

        :param node_func: Function which is applied on all nodes of the tree.
        :return: Return value of the root.
        """
        stack = []

        for node in self.primitives:
            if node.arity == 0:
                stack.append(node_func(node, []))
            else:
                args = stack[-node.arity:]
                stack = stack[:-node.arity]

                stack.append(node_func(node, args))

        if len(stack) > 1:
            raise ValueError("Bad tree")  # TODO specific

        return stack.pop()

    def subtree(self, root_ind):
        """
        Returns the end position of the subtree in the tree individual.

        :param root_ind: Position of the root (index of the beginning).
        :return: The end index of the subtree.
        """
        curr = self.primitives[root_ind]

        arity_rem = curr.arity
        init_h = curr.depth
        max_h = init_h

        while arity_rem > 0:
            root_ind = root_ind - 1
            curr = self.primitives[root_ind]
            max_h = max(max_h, curr.depth)

            arity_rem = arity_rem - 1 + curr.arity

        return root_ind, (max_h - init_h + 1)

    def validate_tree(self):
        """
        Validates the tree, raises an exception if its children are invalid.
        """

        def validate_node(node, child_list):
            if node.arity != len(child_list):
                raise ValueError("Invalid number of children.")  # TODO specific

            child_id = 0
            for in_type in node.node_type[0]:
                for i in range(in_type.arity):
                    child = child_list[child_id + i]
                    if child.node_type[1] != in_type.name:
                        raise ValueError("Invalid child type.")  # TODO specific

                    if child.depth != node.depth + 1:
                        raise ValueError("Invalid child height.")  # TODO specific

                child_id += in_type.arity

            return node

        self.run_tree(validate_node)

        if self.max_height != max(prim.depth + 1 for prim in self.primitives):
            raise ValueError("Invalid tree height.")  # TODO specific


class DeapTreeIndividual(GpTreeIndividual):
    """
    Represents an individual which contains DEAP defined fitness.
    """
    def __init__(self, prim_list, max_height):
        super().__init__(prim_list, max_height)
        self.fitness = DeapTreeIndividual.Fitness()
        self.compiled_pipe = None
        self.test_stats = None

    class Fitness(base.Fitness):
        def __init__(self, values=()):
            self.weights = (1.0, -1.0)
            super().__init__(values)

    def reset(self):
        del self.fitness.values

        self.compiled_pipe = None
        self.test_stats = None


class GpPrimitive:
    """
    Represents a node in the GP tree. It is defined by its type and
    arity values.

    Its name and keyword dictionary hold information about the function
    or object, which is associated with the node.
    """

    def __init__(self, name, obj_kwargs, node_type, arity, depth):
        """
        Creates an instance of a GP tree node. Number of its children
        is specified by its arity, types of children are determined by its
        PrimType type. Its children must be ordered according to the input
        type list.

        :param str name: Name key associated with the node.
        :param dict obj_kwargs: Keyword arguments associated with the node.
        :param (list of PrimType, str) node_type:
            A tuple, where first item is a list of input types with arities
            and the second item is output type name.
        :param int arity: Sum of all arities of types.
        :param int height: Depth of the node.
        """
        self.name = name
        self.obj_kwargs = obj_kwargs
        self.node_type = node_type
        self.arity = arity
        self.depth = depth

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))

        return new

    def __eq__(self, other):
        if not isinstance(other, GpPrimitive):
            return False

        if self.name != other.name:
            return False

        if self.arity != other.arity or self.node_type != other.node_type:
            return False

        if self.obj_kwargs != other.obj_kwargs:
            return False

        return True

    def __repr__(self):
        return 'GpPrimitive(name=' + self.name + ", arity={}".format(self.arity)\
               + ", height={})".format(self.depth)

    @property
    def out_type(self):
        return self.node_type[1]

    class PrimType:
        """
        Represents the input type which must match the output type of
        a certain number of children (determined by ``arity``).
        """
        def __init__(self, name, arity):
            """
            Construct a new instance of the type associated with an arity.

            :param name: Name of the type.
            :param arity: Arity associated with this type.
            """
            self.name = name
            self.arity = arity

        def __eq__(self, other):
            if not isinstance(other, GpPrimitive.PrimType):
                return False

            if self.name != other.name or self.arity != other.arity:
                return False

            return True


class GpTerminalTemplate:
    """
    Represents a leaf of the GP tree.
    Creates a primitive with no inputs. The output type is specified.

    The ``arity`` of the primitive will be set to 0 and ``node_type`` is set to ``([], out_type)``.

    The keyword arguments are chosen from lists of possible values.
    """
    def __init__(self, name, out_type, group=None):
        """
        Creates a new instance of a terminal template.

        :param str name: Name key associated with the node.
        :param str out_type: Name of the node output type.
        """
        self.name = name
        self.type_arities = []
        self.out_type = out_type
        self.group = group

    def create_primitive(self, curr_height, max_arity, kwargs_dict):
        """
        Creates an instance of a ``GpPrimitive`` from the template.

        Selects keyword arguments from ``kwargs_dict``. For every key,
        the dict should contain a list of possible values.

        :param int curr_height: Height at which the node is generated.
        :param int max_arity: Maximum arity of all nodes in the tree, not used.
        :param dict kwargs_dict: Dictionary which contains possible keyword argument values.
        :return: A new instance of GpPrimitive.
        """
        prim_kwargs = _choose_kwargs(kwargs_dict)

        prim_type = ([], self.out_type)

        return GpPrimitive(self.name, prim_kwargs, prim_type, 0, curr_height)


class TypeArity:
    """
    Represents a variable node arity associated with a type.
    """
    def __init__(self, prim_type, arity_range):
        """
        Constructs a new instance of the typed arity.

        :param str prim_type: Type associated with the arity.
        :param int or (int, int) or (int, 'n') arity_range:
            Arity of input nodes, either a constant, or a range (inclusive), or a range
            without a specified upper bound (in the case of (int, 'n')).
        """
        self.prim_type = prim_type
        self.arity_range = arity_range

        if isinstance(self.arity_range, tuple):
            if self.arity_range[1] != 'n' and self.arity_range[0] > self.arity_range[1]:
                raise ValueError("Invalid arity range.")  # TODO specific
        elif isinstance(self.arity_range, int):
            if self.arity_range <= 0:
                raise ValueError("Arity must be greater than 0.")  # TODO specific
        else:
            raise ValueError("Invalid arity type.")  # TODO specific

    def is_valid_arity(self, arity):
        if isinstance(self.arity_range, tuple):
            if arity < self.arity_range[0]:
                return False

            if self.arity_range[1] != 'n' and arity > self.arity_range[1]:
                return False

            return True

        if isinstance(self.arity_range, int):
            return self.arity_range == arity

        raise ValueError("Invalid arity object.")  # TODO specific

    def choose_arity(self, max_arity):
        """
        Chooses an integer arity from the arity range or
        returns the arity in case of a single int.

        :param int max_arity:
            The upper bound of arities. Is applied for variable upper bounds and for
            ranges where the lower bound is less than this parameter.
        :return int: Returns a number from the range of possible arities.
        """
        if not isinstance(self.arity_range, tuple):
            return self.arity_range

        a_from, a_to = self.arity_range

        # is not applied for ranges which are greater than ``max_arity`` as it could break them
        if a_to == 'n' or a_to > max_arity >= a_from:
            a_to = max_arity
            
        return random.randint(a_from, a_to)
        

class GpFunctionTemplate:
    """
    Represents an inner node of the GP tree.
    Creates a primitive which accepts inputs. The output type is also specified.

    The ``arity`` of the primitive is determined by arities associated with types. These
    are ordered (the node children must maintain the same order from left to right). The
    final arities are chosen when a primitive is created from the template.

    The keyword arguments are chosen from lists of possible values.
    """
    def __init__(self, name, type_arities, out_type, group=None):
        """
        Creates a new instance of a function template.

        :param str name: Name key associated with the node.
        :param list of TypeArity type_arities:
            Ordered list of children arity ranges (associated with a type).
        :param str out_type: Name of the output type.
        """
        self.name = name
        self.type_arities = type_arities
        self.out_type = out_type
        self.group = group
        
    def create_primitive(self, curr_height, max_arity, kwargs_dict, in_type=None):
        """
        Creates an instance of a ``GpPrimitive`` from the template.

        Selects keyword arguments from ``kwargs_dict``. For every key,
        the dict should contain a list of possible values.

        Select final arities of every TypeArity in the list. If some of
        the arities result in 0, they are not added into the final input type.

        :param int curr_height: Height at which the node is generated
        :param int max_arity: Maximum arity value which can be chosen for a single TypeArity.
        :param dict kwargs_dict: Dictionary which contains possible keyword argument values.
        :param in_type: Input type; if provided, it is used instead of generating a new random type.

        :return: A new instance of GpPrimitive
        """
        prim_kwargs = _choose_kwargs(kwargs_dict)

        def create_type(t):
            arity = t.choose_arity(max_arity)

            # do not construct a type in case of 0
            if arity == 0:
                return None

            return GpPrimitive.PrimType(t.prim_type, arity)

        if in_type is None:
            # ordered list of final typed arities
            in_type = [create_type(t_a) for t_a in self.type_arities]
            in_type = [t for t in in_type if t is not None]

        # TODO check the provided in_type (and maybe change it a bit)

        # sum of all arities
        arity_sum = functools.reduce(lambda s, t: s + t.arity, in_type, 0)
        
        return GpPrimitive(self.name, prim_kwargs, (in_type, self.out_type), arity_sum, curr_height)


def _choose_kwargs(kwargs_dict):
    """
    Chooses keyword argument values from the argument dictionary, which contains
    list of possible values for every key.

    :param dict kwargs_dict: Dictionary of possible key values.
    :return dict: Dictionary with one value selected per key.
    """
    if kwargs_dict is None:
        return {}

    chosen_kwargs = {}

    for keyword, arg_list in kwargs_dict.items():
        arg = random.choice(arg_list)
        chosen_kwargs[keyword] = arg

    return chosen_kwargs



# -*- coding: utf-8 -*-

"""
This module defines genetic operators used in the evolution as well as the main loop
of the algorithm and initialization methods.
"""

import logging
import random
import time

from deap import tools
from functools import partial, wraps
from itertools import chain
from joblib import Parallel, delayed

from genens.log_utils import set_log_handler
from genens.render.graph import tree_str

from stopit import ThreadingTimeout as Timeout, TimeoutException


def gen_individual(toolbox, config, out_type='out'):
    """
    Generates a random tree individual using the toolbox and arity and height configuration.
    The ``ou_type`` parameter specifies the output type of the root of the tree.

    :param toolbox: Toolbox which contains methods to create the individual.
    :param GenensConfig config: Configuration of Genens
    :param out_type: Output type of the root of the tree individual.
    :return: A new random tree individual.
    """
    arity = random.randint(config.min_arity, config.max_arity)  # randint is inclusive for both limits
    height = random.randint(config.min_height, config.max_height)

    return toolbox.individual(max_height=height, max_arity=arity, first_type=out_type)


@set_log_handler
def gen_valid(toolbox, timeout=100):
    """
    Tries to generate an individual with a valid items. Raises an error if it does
    not succeed in ``timeout`` iterations.

    :param toolbox: Toolbox with initialization and evaluation methods.
    :param timeout: Number of iterations to perform.
    :return: A valid individual.
    """

    for i in range(timeout):
        ind = toolbox.individual()
        score = toolbox.evaluate(ind)

        logger = logging.getLogger("genens")
        logger.debug(f"Generate valid ({i}/{timeout}):\n {tree_str(ind, with_hyperparams=True)}")

        if score is not None:
            ind.fitness.values = score
            return ind

    raise ValueError("Couldn't generate a valid individual.")  # TODO specific


@set_log_handler
def _perform_cx(cx_func, cx_pb, ch1, ch2, **kwargs):
    parent1_str = tree_str(ch1)
    parent2_str = tree_str(ch2)

    if random.random() < cx_pb:
        ch1, ch2 = cx_func(ch1, ch2, **kwargs)
        ch1.reset()
        ch2.reset()

        logger = logging.getLogger("genens")
        logger.debug(f"Crossover ({cx_func}):\n {parent1_str}\n x \n{parent2_str}\n ->"
                     f"\nCh1:\n{tree_str(ch1)}\nCh2:\n{tree_str(ch2)}")

    return ch1, ch2


@set_log_handler
def _perform_mut(mut_func, mut_pb, mut, **kwargs):
    parent_str = tree_str(mut, with_hyperparams=True)

    if random.random() < mut_pb:
        mut = mut_func(mut, **kwargs)
        mut.reset()

        logger = logging.getLogger("genens")
        logger.debug(f"Mutation ({mut_func}):\n {parent_str}\n -> \n{tree_str(mut, with_hyperparams=True)}")

    return mut


@set_log_handler
def _perform_eval(eval_func, ind):
    return eval_func(ind)


def evolution_timeout(fn):
    @wraps(fn)
    def with_timeout(*args, **kwargs):
        if 'timeout' not in kwargs or kwargs['timeout'] is None:
            return fn(*args, **kwargs)

        try:
            timeout = kwargs['timeout']
            with Timeout(timeout, swallow_exc=False):
                fn(*args, **kwargs)
        except TimeoutException:
            logger = logging.getLogger("genens")
            logger.debug(f"Time limit exceeded: {timeout}.")

    return with_timeout


def _get_time_left(start_time, timeout=None):
    if timeout is None:
        return None

    remaining = time.monotonic() - start_time
    return timeout - remaining


@evolution_timeout
def ea_run(population, toolbox, n_gen, pop_size, cx_pb, mut_pb, mut_args_pb, mut_node_pb,
           hc_mut_pb=0.2, hc_n_nodes=3, n_jobs=1, timeout=None,
           min_large_tree_height=3, verbose=1):
    """
    Performs a run of the evolutionary algorithm.

    :param population: Initial population of individuals.
    :param toolbox: Toolbox with methods to use in the evolution.
    :param n_gen: Number of generations.
    :param pop_size: Population size.
    :param cx_pb:
    :param mut_pb:
    :param mut_args_pb:
    :param mut_node_pb:
    :param n_jobs: Number of jobs for multiprocessing
                   ``n_jobs = k, k processors are used
                   ``n_jobs`` = 1, multiprocessing is not used
                   ``n_jobs = k, k + 1 processors are used - for -1, all processors are used

    :param  min_large_tree_height: Perform structural changes only on trees of at least this height.
    :param verbose: Print verbosity
    """
    try:
        start_time = time.monotonic()

        evaluate_func = partial(_perform_eval, toolbox.evaluate, log_setup=toolbox.log_setup)

        if verbose >= 1:
            print('Initial population generated.')

        logger = logging.getLogger("genens")
        for i, ind in enumerate(population):
            logger.debug(f"Generation 0, individual {i}:\n{tree_str(ind, with_hyperparams=True)}")

        # evaluate first gen
        scores = toolbox.map(evaluate_func, population,
                             n_jobs=n_jobs, timeout=_get_time_left(start_time, timeout=timeout))

        for ind, score in zip(population, scores):
            if score is None:
                continue

            ind.fitness.values = score

        if verbose >= 1:
            print("First gen evaluated")

        # remove individuals which threw exceptions and generate new valid individuals
        population[:] = [ind for ind in population if ind.fitness.valid]
        valid = Parallel(n_jobs=n_jobs, timeout=_get_time_left(start_time, timeout=timeout))(
                delayed(gen_valid)(toolbox, log_setup=toolbox.log_setup)
                for _ in range(pop_size - len(population))
        )
        population += valid

        toolbox.update(population, 0)
        population[:] = toolbox.select(population, pop_size)  # assigns crowding distance

        for g in range(n_gen):
            if verbose >= 1:
                print("Gen {}".format(g))

            for i, ind in enumerate(population):
                logger.debug(f"Generation {g}, individual {i}:\n{tree_str(ind, with_hyperparams=True)}")

            # selection for operations
            population[:] = tools.selTournamentDCD(population, pop_size)
            larger_trees = [tree for tree in population if tree.max_height >= min_large_tree_height]
            all_offspring = []

            with Parallel(n_jobs=n_jobs, timeout=_get_time_left(start_time, timeout=timeout)) as parallel:
                if verbose >= 2:
                    print(f"Gen {g} - crossover")

                # crossover - subtree
                offspring = toolbox.map(toolbox.clone, larger_trees)
                offspring = parallel(
                    delayed(_perform_cx)(
                        toolbox.cx_one_point,  # func
                        cx_pb, ch1, ch2, min_node_depth=min_large_tree_height - 1, log_setup=toolbox.log_setup  # args
                    )
                    for ch1, ch2 in zip(offspring[::2], offspring[1::2])
                )
                offspring = list(chain.from_iterable(offspring))  # chain cx tuples to list of offspring
                all_offspring += offspring

                if verbose >= 2:
                    print(f"Gen {g} - mutation")

                # mutation - subtree
                offspring = toolbox.map(toolbox.clone, larger_trees)
                offspring = parallel(
                    delayed(_perform_mut)(
                        toolbox.mutate_subtree,  # func
                        mut_pb, mut, log_setup=toolbox.log_setup  # args
                    )
                    for mut in offspring
                )
                all_offspring += offspring

                # mutation - node swap
                offspring = toolbox.map(toolbox.clone, population)
                offspring = parallel(
                    delayed(_perform_mut)(
                        toolbox.mutate_node_swap,  # func
                        mut_node_pb, mut, log_setup=toolbox.log_setup  # args
                    )
                    for mut in offspring
                )
                all_offspring += offspring

                # for large new offspring, perform hc to make them competitive
                @set_log_handler
                def apply_gradual(offs, mut_pb, n_nodes):
                    if offs.max_height < min_large_tree_height:
                        return offs

                    return _perform_mut(toolbox.gradual_hillclimbing, mut_pb, offs, n_nodes=n_nodes)

                all_offspring = parallel(
                    delayed(apply_gradual)(
                        off, hc_mut_pb, hc_n_nodes, log_setup=toolbox.log_setup
                    )
                    for off in all_offspring
                )

                # ----------------

                # mutation - node args
                offspring = toolbox.map(toolbox.clone, population)
                offspring = parallel(
                    delayed(_perform_mut)(
                        toolbox.mutate_args,  # func
                        mut_args_pb, mut, log_setup=toolbox.log_setup  # args
                    )
                    for mut in offspring
                )
                all_offspring += offspring

                offs_to_eval = [ind for ind in all_offspring if not ind.fitness.valid]

                # evaluation of changed offspring
                scores = toolbox.map(evaluate_func, offs_to_eval,
                                     n_jobs=n_jobs, timeout=_get_time_left(start_time, timeout=timeout))
                for off, score in zip(offs_to_eval, scores):
                    if score is None:
                        continue

                    off.fitness.values = score

                # remove offspring which threw exceptions
                all_offspring[:] = [ind for ind in all_offspring if ind.fitness.valid]
                valid = parallel(delayed(gen_valid)(toolbox, log_setup=toolbox.log_setup)
                                 for _ in range(pop_size - len(all_offspring)))
                all_offspring += valid

            population[:] = toolbox.select(population + all_offspring, pop_size)

            toolbox.update(population, g + 1)

        if verbose >= 1:
            print("Evolution completed")
    except TimeoutError:
        print("Evolution timed out.")



import os
import sys

from . import process_results as pr

if __name__ == "__main__":
    x_axis = sys.argv[1]
    eval_metric = sys.argv[2]
    out_dir = sys.argv[3]

    dirs = sys.argv[4:]

    if not len(dirs):
        sys.exit('Must provide at least one directory')

    dirs = [os.path.normpath(d) for d in dirs]
    columns = [os.path.basename(d) for d in dirs]
    
    df = pr.get_score_stats(columns, dirs)
    
    pr.boxplot_compare_columns(df, out_dir, x_axis=x_axis, eval_metric=eval_metric)
    stats = pr.boxplot_get_stats(df)
    print(stats.to_latex())




import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="genens",
    version="0.1.9",
    author="Gabriela Suchoparova",
    author_email="gabi.suchoparova@gmail.com",
    description="A genetic AutoML system for ensemble methods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gabrielasuchopar/genens",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        '': ['.logging_config.json']
    },
    python_requires='>=3.6',
    install_requires=[
        'deap',
        'graphviz',
        'joblib',
        'openml',
        'matplotlib',
        'numpy',
        'pygraphviz',
        'scikit-learn>=0.22',
        'seaborn',
        'stopit'
    ]
)



import logging
import math
import random
from functools import partial

from genens.gp.tree import swap_subtrees
from genens.gp.types import GpFunctionTemplate, GpTreeIndividual


def mutate_subtree(toolbox, gp_tree, eps=0.2, min_node_depth=1):
    """
    Replaces a randomly chosen subtree with a new random tree. The height of the generated subtree
    is between 1 and previous subtree height + ``eps``.

    :param toolbox: Toolbox of the genetic algorithm.
    :param GpTreeIndividual gp_tree:
    :param float eps:
        Height of the new subtree lies in the interval ((1-eps) * old_height, (1+eps) * old_height).

    :param int min_node_depth: Minimal mutation point depth in both trees.

    :return: The mutated tree.
    """
    # mutation does not replace the whole tree
    if gp_tree.max_height < 2:
        return gp_tree

    # select a node other than the root
    eligible_inds = [ind for ind, node in enumerate(gp_tree.primitives) if node.depth >= min_node_depth]
    if not len(eligible_inds):
        logger = logging.getLogger("genens")
        logger.debug(f"Mutation (subtree) - tree too small:\n{gp_tree}")
        return gp_tree

    mut_end_point = random.choice(eligible_inds)
    _, subtree_height = gp_tree.subtree(mut_end_point)

    lower = math.floor((1.0 - eps) * subtree_height)
    upper = math.ceil((1.0 + eps) * subtree_height)

    new_height = random.randint(lower, upper + 1)  # generate a smaller or a bigger subtree
    new_tree = toolbox.individual(max_height=new_height,
                                  first_type=gp_tree.primitives[mut_end_point].out_type)
    new_root_point = len(new_tree.primitives) - 1

    offs = swap_subtrees(gp_tree, new_tree, mut_end_point, new_root_point, keep_2=False)

    return offs[0]


def _node_types_match(node, node_template):
    """
    Verifies whether input and output types of two nodes match. The first has fixed arities of
    the input type whereas the second is a template with variable arities.

    :param node: Node with fixed input types arities.
    :param node_template: A node template with variable arities.

    :return bool: True if the types of the nodes match.
    """
    # out types match?
    if node.node_type[1] != node_template.out_type:
        return False

    # in types (and arities) match?
    in_types = node.node_type[0]
    template_types = node_template.type_arities

    if len(in_types) != len(template_types):
        return False

    for in_type_1, in_type_2 in zip(in_types, template_types):
        # check type compatibility
        if in_type_1.name != in_type_2.prim_type:
            return False

        # check arity compatibility
        if not in_type_2.is_valid_arity(in_type_1.arity):
            return False

    return True


def mutate_node_swap(config, gp_tree):
    """
    Mutates randomly a node of the tree --- replaces it with a new node with the same type.

    :param GenensConfig config: Configuration which contains node definitions.
    :param GpTreeIndividual gp_tree: Individual to be mutated
    :return: Mutated individual.
    """

    swap_ind = random.randrange(len(gp_tree.primitives))
    swap_node = gp_tree.primitives[swap_ind]

    out_type = swap_node.node_type[1]

    possible_templates = [tm for tm in config.full_config[out_type]
                          if _node_types_match(swap_node, tm)]

    possible_templates += [tm for tm in config.term_config[out_type]
                           if _node_types_match(swap_node, tm)]

    chosen_tm = random.choice(possible_templates)

    if isinstance(chosen_tm, GpFunctionTemplate):
        new_node = chosen_tm.create_primitive(swap_node.depth, config.max_arity,
                                              config.kwargs_config[chosen_tm.name],
                                              in_type=swap_node.node_type[0])
    else:
        new_node = chosen_tm.create_primitive(swap_node.depth, config.max_arity,
                                              config.kwargs_config[chosen_tm.name])

    gp_tree.primitives[swap_ind] = new_node
    return gp_tree


def mutate_args(config, gp_tree, multiple_nodes=False, multiple_args=False):
    _, mut_nodes = _get_nodes_for_mut(gp_tree, multiple_nodes=multiple_nodes)
    if not len(mut_nodes):
        return gp_tree

    for node in mut_nodes:
        _mutate_node_args(config, node, multiple=multiple_args)

    return gp_tree


def _get_nodes_for_mut(gp_tree, multiple_nodes=False):
    prim_inds = [i for i, prim in enumerate(gp_tree.primitives) if len(prim.obj_kwargs)]
    if not len(prim_inds):
        return [], []

    if multiple_nodes:
        n_nodes = random.randint(1, len(prim_inds))
    else:
        n_nodes = 1

    ind_sample = random.sample(prim_inds, n_nodes)
    return ind_sample, [gp_tree.primitives[i] for i in ind_sample]


def _mutate_node_args(config, mut_node, multiple=False):
    # no parameters to mutate
    if not len(mut_node.obj_kwargs):
        return mut_node

    all_keys = list(mut_node.obj_kwargs.keys())

    if multiple:
        n_kwargs = random.randint(1, len(all_keys))
        mut_kwargs = random.sample(all_keys, n_kwargs)

        for key in mut_kwargs:
            _mut_arg(config, mut_node, key)
    else:
        mut_arg = random.choice(all_keys)
        _mut_arg(config, mut_node, mut_arg)

    return mut_node


def mutate_gradual_hillclimbing(toolbox, config, gp_tree: GpTreeIndividual, multiple_args=True,
                                hc_repeat=5, n_nodes=3, keep_last=False):
    def mut_ind(ind, gp_tree):
        _mutate_node_args(config, gp_tree.primitives[ind], multiple=multiple_args)

    mut_inds, _ = _get_nodes_for_mut(gp_tree, multiple_nodes=True)
    if not len(mut_inds):
        return gp_tree

    n_nodes = min(n_nodes, len(mut_inds))
    mut_inds = random.sample(mut_inds, n_nodes)
    mut_inds = random.sample(mut_inds, n_nodes)

    for i in mut_inds:
        mut_func = partial(mut_ind, i)
        gp_tree = perform_hillclimbing(toolbox, gp_tree, mut_func, hc_repeat=hc_repeat, keep_last=keep_last)

    return gp_tree


def perform_hillclimbing(toolbox, gp_tree, mut_func,
                         hc_repeat=1, keep_last=False):
    # hill-climbing initial fitness
    if not gp_tree.fitness.valid:
        score = toolbox.evaluate(gp_tree)

        # mutate only valid individuals
        if score is None:
            return gp_tree
        gp_tree.fitness.values = score

    logger = logging.getLogger("genens")

    # hill-climbing procedure
    has_mutated = False
    for i in range(hc_repeat):
        mutant = toolbox.clone(gp_tree)
        mut_func(mutant)

        score = toolbox.evaluate(mutant)
        # skip invalid mutants
        if score is None:
            continue

        # return the last one if no mutant was better than the first individual
        should_keep_last = keep_last and i == hc_repeat - 1 and not has_mutated

        # the mutant is better, keep it
        if score >= gp_tree.fitness.values or should_keep_last:
            logger.debug(f"Hillclimbing {i}/{hc_repeat}: score {gp_tree.fitness.values} -> {score},"
                         f"\n{gp_tree}\n>\n{mutant}")

            gp_tree = mutant
            has_mutated = True

    return gp_tree


def _mut_arg(config, node, key):
    """
    Mutates a keyword argument of a single node.

    :param GenensConfig config: Configuration of Genens which contains lists of possible keyword arguments values.
    :param node: Node to be mutated.
    :param key: Key of the argument which will be mutated
    """

    node.obj_kwargs[key] = random.choice(config.kwargs_config[node.name][key])



from genens import GenensClassifier
from genens.workflow.evaluate import SampleCrossValEvaluator
from genens.gp.crossover import crossover_one_point

if __name__ == "__main__":
    #X, y = load_dataset('wilt')

    eval = SampleCrossValEvaluator(cv_k=5, per_gen=True)
    clf = GenensClassifier(n_jobs=1, pop_size=12, n_gen=2, evaluator=eval)
    
    #clf.fit(X, y)
    pop = clf._population

    print(pop)

    for tree1, _ in zip(pop[1:], pop[:-1]):
        tree1, tree2 = crossover_one_point(tree1, tree1)
        #mut1 = mutate_subtree(clf._toolbox, tree1)

    pass